#
# Class to support PyXspec usage
#
# David Homan 10.08.2022
#

import xspec
import numpy as np

from itertools import compress

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.ticker import FormatStrFormatter
plt.rc('font', family='serif')


class PlotXspec:
    '''
    Provides basic plotting and inspection functions when using PyXspec.
    ---
    The following public methods are defined:
    
    first_look:
    A quick first look at the data
    
    plot_model_and_data:
    Plot the data as well as the currently active model
    
    plot_chisq_contours:
    Plot 2D steppar results, including several chi-squared contours
    
    calc_error_from_1Dchisq:
    Plot 1D steppar results and estimate a single parameter's uncertainty
    based on a specified chi-squared value
   
    print_model_results
    Print fitting results to screen
    
    print errors
    Calculate errors on model parameters (for a specified chi-squared
    value) and print to screen
    '''
    
    def __init__(self):
        return None


    #####################################
    #                                   #
    #      (0) Helper Functions         #
    #                                   #
    #####################################

    def _get_xspec_data(self,model=True,plottype='folded',
                        idsp=1,idmd=1,model_name=None,
                        backgrnd=False,
                        rebinsig=None,rebinbnum=None):
        '''
        Get data from XSPEC plot
        
        Keyword arguments:
        model      ---  boolean; True if model is included in xspec
                        plot (default is True)
        plottype   ---  string; 'folded' or 'unfolded' setting the main
                        plot type (and units of the y-axis)
        idsp       ---  int; ID of spectrum to plot
                        (in plotGroup; default is 1)
        idmd       ---  int; ID of model data group to include in plot 
                        (in AllModels; default is 1)
        model_name ---  string or list of strings; named models to
                        include in plot. This is of particular use
                        when there are multiple named models that
                        contribute to the total model (i.e. when 
                        there are multiple sources defined in XSpec
                        for the same spectrum). The data for the 
                        models will be returned under 'mc'. Default 
                        is None; in this case the method will check
                        for additive model components
        backgrnd   ---  boolean; return background data (note: this
                        is always in units of counts/s/keV, regardless
                        of plttype)
        rebinsig   ---  int; minimum sigma per bin (plotting only)
        rebinbnum  ---  int; maximum number of bins to combine when
                        rebinning using rebinsig (plotting only)
                         
        Returns:
        x   ---  list of floats; energy bins
        y   ---  list of floats; counts/flux
        xe  ---  list of floats; width of bins
        ye  ---  list of floats; error on counts/flux
        pm  ---  list of floats; model values (if model==True)
        rs  ---  NumPy array of floats; residuals (if model==True)
        rt  ---  NumPy array of floats; error-normalised residuals 
                                        (if model==True)
        mc  ---  list of lists of floats; model component values
                                         (if model==True)
        bg  ---  list of floats; counts/flux of the background
                                 (if backgrnd==True)
        '''
        # Initial setup and xspec plot
        xspec.Plot.xAxis = "keV"
        if rebinsig:
            assert rebinbnum is not None
            xspec.Plot.setRebin(minSig=rebinsig,
                                maxBins=rebinbnum,
                                groupNum=-1)
        else:
            xspec.Plot.setRebin(minSig=0,maxBins=1,groupNum=-1)
        if plottype == 'folded':
            plttype = 'data'
        elif plottype == 'counts':
            plttype = 'counts'
        elif plottype == 'unfolded':
            plttype = 'euf'
        else: raise ValueError
        xspec.Plot(plttype)
        # Get coordinates from plot:
        x  = xspec.Plot.x(plotGroup=idsp)
        xe = xspec.Plot.xErr(plotGroup=idsp)
        y  = xspec.Plot.y(plotGroup=idsp)
        ye = xspec.Plot.yErr(plotGroup=idsp)
        results = [x,y,xe,ye]
        if model:
            # get model
            pm = xspec.Plot.model(plotGroup=idsp)
            # calculate residuals
            rs = np.array(y)-np.array(pm)
            ye_corr = [1 if v==0 else v for v in ye] # erroneous errors can occur
            rt = rs/np.array(ye_corr)
            mc = []
            if model_name is not None:
                for m in model_name:
                    mc += [xspec.AllModels(idmd,m).folded(idsp)]
                # the folded fluxes need to be adjusted such
                # that the combined flux matches that of the
                # model flux. This is a stop-gap measure, as
                # we cannot access the plotted fluxes for the
                # named models directly if there are multiple
                # models for the same source
                mtot    = [sum(mx) for mx in zip(*mc)]
                ratio_m = [mf/mcf for mcf,mf in zip(mtot,pm)]
                mc = [[mx*r for mx,r in zip(mcl,ratio_m)] for mcl in mc]
            else:
                # check for additive components in the model
                ## additive components are not included when plotting
                ## 'data' or 'counts' => if not present yet, plot 'euf'
                if xspec.Plot.nAddComps() == 0 and plttype != 'euf':
                    xspec.Plot('euf')
                if xspec.Plot.nAddComps() > 0:
                    xspec.Plot.add = True
                    xspec.Plot(plttype)  # redraw plot to include components
                    for ii in range(xspec.Plot.nAddComps()):
                        mc += [xspec.Plot.addComp(ii+1,plotGroup=idsp)]
            results += [pm,rs,rt,mc]
        if backgrnd:
            xspec.Plot('back')
            bg  = xspec.Plot.y(plotGroup=idsp)
            bge = xspec.Plot.yErr(plotGroup=idsp)
            results += [bg,bge]
        # reset rebinning if necessary
        if rebinsig:
            xspec.Plot.setRebin(minSig=0,maxBins=1,groupNum=-1)
        return results

    def _extrapolate_model_log(self,x,xw,y):
        '''
        Extend the plotted model (y), such that when displayed
        it extends the full range spanned by the data 
        (including the edges of the bins -x-).
        
        NOTE: this functionality is for display only.
        
        Positional arguments:
        x  --- list of floats; xbins
        xw --- list of floats; width of xbins
        y  --- list of floats; model values (belonging to xbins)
        
        Returns:
        x  --- list of floats; extended xbins, expanded at start and end
        y  --- list of floats; extended model values
        '''
        # lower range
        dx = np.log10(x[1]/x[0])
        x  = [ x[0]-xw[0] ] + x
        if y[0]!=0:
            dy = np.log10(y[1]/y[0])
            dlogx = np.log10(x[0]/x[1])
            y  = [ y[0] * np.power(10,dlogx*(dy/dx)) ] + y
        else:
            y  = [ y[0] ] + y
        # upper range
        dx = np.log10(x[-1]/x[-2])
        x  = x + [ x[-1]+xw[-1] ]
        if y[-2]!=0:
            dy = np.log10(y[-1]/y[-2])
            dlogx = np.log10(x[-2]/x[-1])
            y  = y + [ y[-1] * np.power(10,-dlogx*(dy/dx)) ]
        else:
            y  = y + [ y[-1] ]
        return (x,y)

    def _find_loaded_data_format(self):
        '''
        Find the format of the currently loaded XSPEC
        data and return the pertinent values in a list
        of dictionaries, one entry per loaded spectrum.
        
        Positional arguments:
        None
        
        Returns:
        orig_data  list of dicts; load settings for
                   each loaded spectrum
        '''
        orig_data = []
        for ii in range(1,xspec.AllData.nSpectra+1):
            sp  = xspec.AllData(ii).fileName
            try:
                bkg = xspec.AllData(ii).background.fileName
            except:
                bkg = ''
            try:
                rmf = xspec.AllData(ii).response.rmf
            except:
                rmf = ''
            try:
                arf = xspec.AllData(ii).response.arf
            except:
                arf = ''
            exp = xspec.AllData(ii).exposure
            ign = xspec.AllData(ii).ignoredString()
            dg  = xspec.AllData(ii).dataGroup
            ind = xspec.AllData(ii).index
            orig_data.append({'sp':sp,'bkg':bkg,
                              'rmf':rmf,'arf':arf,
                              'exp':exp,'ign':ign,
                              'dg':dg,'ind':ind})
        return orig_data

    def _find_var_params_model(self,model):
        '''
        find the variable parameters of a given XSPEC model
        
        Positional arguments:
        model   --- XSPEC model
        
        Returns:
        cmp_var --- dict; keys are model component names, items are
                    lists of the variable parameters per component
        '''
        cmp_var = {}
        # components of model
        cmp = model.componentNames
        for c in cmp:
            # parameters of components
            params = model.__dict__[c].parameterNames
            isvar  = []
            for p in params:
                if model.__dict__[c].__dict__[p].link == '':
                    isvar += [not model.__dict__[c].__dict__[p].frozen]
                else:
                    isvar += [False]
            # store the variable parameters for the given component
            cmp_var[c] = list(compress(params,isvar))
        return cmp_var

  	
    #################################
    #                               #
    #     (1) Plot the spectra      #
    #                               #
    #################################
        
    def first_look(self,idsp=None,
                   backgrnd=False,
                   ymin=None,ymax=None,
                   xlog=True,ylog=True,
                   rebinsig=None,rebinbnum=10,
                   savename=None):
        '''
        First look at the data (quick and simple).
        
        Keyword arguments:
        idsp              --- int or list of ints; ID of spectra to plot 
                              (in plotGroup; default is all loaded spectra)
        backgrnd          --- boolean; show background data
        ymin              --- float; the lower limit on the y-axis
        ymax              --- float; the top limit on the y-axis
        xlog              --- bool; set logscale on xaxis (default = True)
        ylog              --- bool; set logscale on yaxis (default = True)
        rebinsig          --- int; minimum sigma per bin (plotting only)
        rebinbnum         --- int; maximum number of bins to combine when
                              rebinning (plotting only)
        savename          --- str; name of output file
        
        Returns:
        None
        '''
        colors = ['k']+[plt.cm.Set1(ii) for ii in range(10)]
        if idsp is None:
            idsp = [ii+1 for ii in range(xspec.AllData.nSpectra)]
        elif type(idsp) is int:
            idsp = [idsp]
        fig,ax = plt.subplots(1,figsize=(12,8))
        for ii,idsp_ii in enumerate(idsp):
            # get data
            if backgrnd:
                bins,_,binw,_,rates,ratese = self._get_xspec_data(model=False,
                                                                  idsp=idsp_ii,
                                                                  backgrnd=True,
                                                                  rebinsig=rebinsig,
                                                                  rebinbnum=rebinbnum)
            else:
                bins,rates,binw,ratese = self._get_xspec_data(model=False,
                                                              idsp=idsp_ii,
                                                              rebinsig=rebinsig,
                                                              rebinbnum=rebinbnum)
            # plot using matplotlib
            self._make_spec_plot_main(ax,
                                      bins,rates,binw,ratese,
                                      ymin=ymin,ymax=ymax,
                                      xlog=xlog,ylog=ylog,
                                      clr=colors[ii])
        # save plot if necesarry
        if savename:
            plt.tight_layout()
            plt.savefig(savename)
        else:
            plt.show()

            
    def plot_model_and_data(self,
                            ymin=None,ymax=None,
                            subplot='ratio',plottype='folded',
                            idsp=None,idmd=None,model_name=None,
                            msize=3,colors=None,
                            addxminorticks=None,
                            setxminorticks=None,
                            rebinsig=None,rebinbnum=10,
                            extend=False,
                            savename=None):
        '''
        Create a plot of the model, data, and residuals.
        
        Keyword arguments:
        ymin           --- float; the lower limit on the y-axis
        ymax           --- float; the top limit on the y-axis
        subplot        --- string; 'res' or 'ratio', setting the content of
                           the second plot
        plottype       --- string; 'folded' (data), 'unfolded' (euf), or
                           'counts'; setting the main plot type. Default
                           is 'folded' (counts s^-1 keV^-1)
        idsp           --- int or list of ints; ID of spectra to plot 
                           (in plotGroup; default is all loaded spectra)
        idmd           --- int or list of ints; ID of model to plot (in
                           AllModels; default is the models matching the 
                           spectra defined with idsp). Note: the number of
                           models must match the number of spectra to be 
                           plotted.
        model_name     --- string or list of strings; named models to
                           include in plot. This is of particular use
                           when there are multiple named models that
                           contribute to the total model (i.e. when 
                           there are multiple sources defined in XSpec
                           for the same spectrum). The data for the 
                           models will be returned under 'mc'. Default 
                           is None; in this case the method will check
                           for additive model components.
        addxminorticks --- list; additional minor ticks for x-axis
        setxminorticks --- list; explicitly set minor ticks for x-axis
        rebinsig       --- int; minimum sigma per bin (plotting only)
        rebinbnum      --- int; maximum number of bins to combine when
                           rebinning (plotting only)
        msize          --- int; markersize
        colors         --- list of matplotlib colours to use for the plotted
                           spectra (default is None -> automatic)
        extend         --- bool; set to True to extend the plotted 
                           model to the edges of the plot using
                           a simple interpolation method
                           (defaut is False)
        savename       --- string; name of output file (default is None)
        
        Returns:
        None
        '''
        # check input
        if idsp is None:
            idsp = [ii+1 for ii in range(xspec.AllData.nSpectra)]
        elif type(idsp) is int:
            idsp = [idsp]
        if idmd is None:
            idmd = idsp
        elif type(idmd) is int:
            idmd = [idmd]
        assert len(idsp)==len(idmd), 'the number of models must match the number of spectra'
        if model_name is not None and type(model_name) is not list:
            model_name = [model_name]
        # define colors to use
        if colors is None:
            colors  = ['k']+[plt.cm.Set1(ii) for ii in range(len(idsp)-1)]
        if len(idsp)>1:
            mcolors = colors
        else:
            mcolors = ['r']
        # set up figure
        fig = plt.figure(constrained_layout=False,figsize=(15,12))
        gs = fig.add_gridspec(nrows=7, ncols=1, wspace=0, hspace=0)
        ax = fig.add_subplot(gs[0:5, :])
        ax_res = fig.add_subplot(gs[5:, :])
        for ii,idsp_ii in enumerate(idsp):
            # get data from XSPEC plot
            xdata = self._get_xspec_data(idsp=idsp_ii,idmd=idmd[ii],
                                         model_name=model_name,
                                         plottype=plottype,
                                         rebinsig=rebinsig,rebinbnum=rebinbnum)
            bins,rates,binw,ratese,mplot,res,ratio,mplot_cmp = xdata
            # specify data to be used in subplot
            if subplot=='res': 
                sp_data = res
                sp_datae = ratese
            elif subplot=='ratio':
                sp_data = ratio
                sp_datae = np.ones(len(sp_data))
            else: raise ValueError
            # main plot
            self._make_spec_plot_main(ax,
                                      bins,rates,binw,ratese,
                                      mplot=mplot,mplot_cmp=mplot_cmp,
                                      model_name=model_name,
                                      ymin=ymin,ymax=ymax,
                                      xlog=True,ylog=True,
                                      plottype=plottype,msize=msize,
                                      clr=colors[ii],mclr=mcolors[ii],
                                      addxminorticks=addxminorticks,
                                      setxminorticks=setxminorticks,
                                      xlabel=False,
                                      extend=extend)
            # residuals plot
            self._make_spec_plot_residuals(ax_res,
                                           bins,sp_data,binw,sp_datae,
                                           xlog=True,ylog=False,
                                           clr=colors[ii],
                                           subplot=subplot,msize=msize,
                                           addxminorticks=addxminorticks,
                                           setxminorticks=setxminorticks)
        # save plot if necesarry
        if savename:
            plt.tight_layout()
            plt.savefig(savename)
        else:
            plt.show()

    def _make_spec_plot_basic(self,ax,
                              x,y,xerr,yerr,
                              ymin=None,ymax=None,
                              xlog=True,ylog=True,
                              msize=2,clr='k',
                              addxminorticks=None,
                              setxminorticks=None):
        '''
        Create a basic plot of the spectral data,
        including formatting of the axes.
        '''
        # plot data
        ax.errorbar(x,y,xerr=xerr,yerr=yerr,ls='',
                    c=clr,markersize=msize,marker='s')
        # format axes
        ax.grid()
        if xlog:
            ax.set_xscale('log')
        if ylog:
            ax.set_yscale('log')
        # x-range
        ax.set_xlim((np.amin(x)-xerr[0]/2,np.amax(x)+xerr[-1]/2))
        # y-range
        if not ymin:
            ymin = np.amin(y)*2
            if ylog and ymin<0:
                ymin = 1e-2
        if not ymax:
            ymax = np.amax(y)*2
        ax.set_ylim((ymin,ymax))
        # set and format tick labels
        if addxminorticks is not None:
            minticks = list(ax.get_xticks(minor=True)) + list(addxminorticks)
            minticks.sort()
            xmin,xmax = ax.get_xlim()
            minticks = [m for m in minticks if m > xmin and m < xmax]
            ax.set_xticks(minticks,minor=True)
        elif setxminorticks is not None:
            ax.set_xticks(setxminorticks,minor=True)
        plt.setp(ax.get_xticklabels(minor=True), fontsize=13)#, fontweight='bold')
        plt.setp(ax.get_xticklabels(), fontsize=12, fontweight='bold')
        plt.setp(ax.get_yticklabels(), fontsize=12, fontweight='bold')
        ax.tick_params(axis='y', which='minor',
                       direction='out', length=4, width=1.5)

    def _make_spec_plot_main(self,ax,
                             x,y,xerr,yerr,
                             mplot=None,mplot_cmp=None,
                             model_name=None,
                             ymin=None,ymax=None,
                             xlog=True,ylog=True,
                             plottype='folded',
                             msize=2,
                             clr='k',mclr='r',
                             xlabel=True,
                             addxminorticks=None,
                             setxminorticks=None,
                             extend=False):
        '''
        Create spectral plot based on XSPEC data, including model data,
        if available.
        '''
        # generate basic plot
        self._make_spec_plot_basic(ax,x,y,xerr,yerr,
                                   ymin=ymin,ymax=ymax,
                                   xlog=xlog,ylog=ylog,
                                   msize=msize,
                                   clr=clr,
                                   addxminorticks=addxminorticks,
                                   setxminorticks=setxminorticks)
        # plot model
        if mplot is not None:
            if extend:
                x,mplot = self._extrapolate_model_log(x,xerr,mplot)
            e = [v-err for v,err in zip(x,xerr)] + [x[-1]+xerr[-1]]
            ax.stairs(mplot,edges=e,color=mclr,lw=2,label='total')
        # plot model components
        if mplot_cmp is not None:
            if clr==mclr:
                cmpclr = [mclr]*len(mplot_cmp)
            else:
                cmpclr= [plt.cm.Set1(ii+1) for ii in range(len(mplot_cmp))]
            for ii,(cmp,col) in enumerate(zip(mplot_cmp,cmpclr)):
                if extend:
                    x,mplot = self._extrapolate_model_log(bins,binw,cmp)
                if model_name is not None:
                    e = [v-err for v,err in zip(x,xerr)] + [x[-1]+xerr[-1]]
                    ax.stairs(cmp,edges=e,color=col,lw=1.5,ls='-',label=model_name[ii])
                else:
                    ax.plot(x,cmp,c=col,lw=3,ls=':')
        # set axes labels
        if xlabel:
            ax.set_xlabel('Energy (keV)',fontweight='bold',fontsize=20)
        if plottype == 'folded':
            yl = r'Counts s$^{-1}$ keV$^{-1}$'
        elif plottype == 'counts':
            yl = r'Counts keV$^{-1}$'
        elif plottype == 'unfolded':
            yl = 'keV\n'+r'[Photons cm$^{-2}$ s$^{-1}$ keV$^{-1}$]'
        if model_name is not None:
            ax.legend(loc='best',fontsize=15)
        ax.set_ylabel(yl,fontweight='bold',fontsize=20)
        
    def _make_spec_plot_residuals(self,ax,
                                  x,y,xerr,yerr,
                                  xlog=True,ylog=False,
                                  subplot='res',
                                  msize=2,clr='k',
                                  addxminorticks=None,
                                  setxminorticks=None,
                                  zeroline=True):
        '''
        Create spectral plot of the residuals (used in a 
        subplot of the plot_model_and_data() method).
        '''
        # define y-limits
        ymin = np.amin(y) - yerr[np.argmin(y)]
        ymax = np.amax(y) + yerr[np.argmax(y)]
        # generate basic plot
        self._make_spec_plot_basic(ax,x,y,xerr,yerr,
                                   ymin=ymin,ymax=ymax,
                                   xlog=xlog,ylog=ylog,
                                   msize=msize,clr=clr,
                                   addxminorticks=addxminorticks,
                                   setxminorticks=setxminorticks)
        # mark zero
        if zeroline:
            xlim = ax.get_xlim()
            ax.hlines(0,xlim[0],xlim[1],colors='green')
        # set x-label
        ax.set_xlabel('Energy (keV)',fontweight='bold',fontsize=20)
        # set y-label
        if subplot == 'res':
            yl = 'Residuals'
        else:
            yl = 'Residuals\n'+r'($\sigma$-Norm.)'#alised)'
        ax.set_ylabel(yl,fontweight='bold',fontsize=20)
        # set and format ticks
        ax.tick_params(axis='x', which='minor',
                       direction='out', length=5, width=2)
        ax.xaxis.set_major_formatter(FormatStrFormatter('%.1f'))
        ax.xaxis.set_minor_formatter(FormatStrFormatter('%.2f'))


    #####################################
    #                                   #
    # (2) Print XSPEC results to screen #
    #                                   #
    #####################################
        
        
    def print_model_results(self,model_name=['']):
        '''
        Find the model results (incl. the error) and 
        print to screen. This will be done for all active
        models. The format of the output is as follows
        
        parameter_index  model_component  component_parameter  value  estimated_sigma  unit
        
        keyword argument:
        model_name   list of strings; named models. Default is to assume an unnamed model
        
        Returns:
        None
        '''
        nMod = xspec.AllData.nGroups
        for ii in range(nMod):
            for mname in model_name:
                try:
                    mod = xspec.AllModels(ii+1,mname)
                except:
                    print('Could not load model, perhaps named models should be used?')
                    return
                print(f'\nComponents of model {mname} for data group {ii+1}:')
                for cmp in mod.componentNames:
                    stid = mod.startParIndex-1
                    pars = mod.__dict__[cmp].parameterNames
                    for par in pars:
                        ind = mod.__dict__[cmp].__dict__[par]._Parameter__index + stid
                        val = mod.__dict__[cmp].__dict__[par].values[0]
                        sig = mod.__dict__[cmp].__dict__[par].sigma
                        unit = mod.__dict__[cmp].__dict__[par].unit
                        print ('{}{}{}{:.5e} +/- {:.5e} {}'.format(
                                    str(ind).ljust(4),
                                    cmp.ljust(18),
                                    par.ljust(14),
                                    val,
                                    sig,
                                    unit.rjust(6))
                              )
            print('---')
            method = xspec.Fit.statMethod
            stat = xspec.Fit.statistic
            print('Fit statistic: {}'.format(method))
            print('Test statistic: {:5.2f}'.format(stat))
            if method == 'chi':
                dof = xspec.Fit.dof
                print('Reduced chi-squared: {:5.2f}/{} = {:.3}'
                          .format(stat,dof,stat/dof))

    def print_errors(self,chisq,model_name=['']):
        '''
        Print the errors to screen, using the Xspec 
        error routine. The output has the following format:
        
        parameter_index  model_component  component_parameter  value lower_error  upper_error
        
        positional argument:
        chisq      --- float; chi-squared level at which to calculate the errors. This is
                       the argument passed to XSpec's 'error' function
        model_name --- list of strings; named models. Default is to assume an unnamed model
                    
        Returns:
        None
        '''
        # temporarily switch off the printing to CL by XSpec
        xchat = xspec.Xset.chatter, xspec.Xset.logChatter
        xspec.Xset.chatter    = 0
        xspec.Xset.logChatter = 0
        # calculate errors
        nMod = xspec.AllData.nGroups
        for ii in range(nMod):
            for mname in model_name:
                try:
                    mod = xspec.AllModels(ii+1,mname)
                except:
                    print('Could not load model, perhaps named models should be used?')
                    return
                print(f'\nComponents of model {mname} for data group {ii+1}:')
                try:
                    for cmp in mod.componentNames:
                        stid = mod.startParIndex-1
                        pars = mod.__dict__[cmp].parameterNames
                        for par in pars:
                            pind = mod.__dict__[cmp].__dict__[par]._Parameter__index + stid
                            val = mod.__dict__[cmp].__dict__[par].values[0]
                            xspec.Fit.error('{} {}'.format(chisq,pind))
                            ev = list(mod.__dict__[cmp].__dict__[par].error)
                            if ev[0]!=0:
                                ev[0] = val-ev[0]
                                ev[1] = ev[1]-val
                            print('{} {} {} {:.3e} (-){:.3e} (+){:.3e}'.format(
                                            str(pind).ljust(3),
                                            cmp.ljust(10),
                                            par.ljust(10),
                                            val,
                                            ev[0],
                                            ev[1])
                                 )
                except(Exception):
                        print('Unable to calculate errors, perhaps chisq is too high?\n')
                        xspec.Xset.chatter, xspec.Xset.logChatter = xchat
        # set XSpec's outbut levels back to their original values
        xspec.Xset.chatter, xspec.Xset.logChatter = xchat


    #####################################
    #                                   #
    #  (3) Plot XSPEC chi^2 countours   #
    #                                   #
    #####################################

    def plot_chisq_contours(self,levels=[2.3,4.61,9.21],
                            savename=None):
        '''
        Create a plot of the reduced chi-squared over the
        parameter space set in a steppar run.
        
        Keyword arguments:
        levels    --- list of floats; Delta-chisq levels to include in the
                      plot as contours (default is [2.3,4.61,9.21])
        savename  --- string; filename to store results (default is None)
         
        Returns:
        None
        '''
        chisq = xspec.Fit.statistic
        xspec.Plot('contour')
        x, y = np.meshgrid(xspec.Plot.x(), xspec.Plot.y())
        z = xspec.Plot.z()
        fig, ax = plt.subplots(1,figsize=(10,6))
        z_min, z_max = np.array(z).min(), np.array(z).max()
        c = ax.pcolormesh(x, y, z, cmap='RdBu', 
                          vmin=z_min, vmax=z_max,shading='auto')
        fig.colorbar(c, ax=ax)
        labl = xspec.Plot.labels()
        ax.set_xlabel(labl[0],fontsize=16,fontweight='bold')
        ax.set_ylabel(labl[1],fontsize=16,fontweight='bold')
        ax.set_title(labl[2],fontsize=16,fontweight='bold')
        con = [l+chisq for l in levels]
        if np.amin(z) < con[0]:
            ax.contour(x, y, z, levels=con)
        if savename:
            plt.savefig(savename)
        else:
            plt.show()
            
    def calc_error_from_1Dchisq(self,par,level=1,
                                savename=None):
        '''
        Create plot and calculate relevant uncertainties
        based on 1D chi-squared contour
        
        Positional arguments:
        par       --- the model parameter in question
        
        Keyword arguments:
        level     --- float; Delta-chisq level used to calculate the uncertainties
        savename  --- string; filename to store results (default is None)
        
        Returns:
        None
        '''
        xspec.Plot('contour')
        x, y = xspec.Plot.x(), xspec.Plot.y()
        chisq = xspec.Fit.statistic
        val = par.values[0]
        # Plot
        fig, ax = plt.subplots(1,figsize=(10,6))
        ax.plot(x,y,color='k')
        ax.hlines(chisq+level,np.amin(x),np.amax(y),colors='r',
                  label=r'$\Delta\chi^2$='+f'{level}')
        ax.vlines(val,chisq,np.amax(y),color='k',ls=':',
                  label=f'best fit: {par.name}={val:.3f}')
        # Plot details
        labl = xspec.Plot.labels()
        ax.set_xlabel(labl[0],fontsize=16,fontweight='bold')
        ax.set_ylabel(labl[1],fontsize=16,fontweight='bold')
        ax.set_title(labl[2],fontsize=16,fontweight='bold')
        ax.set_xlim((np.amin(x),np.amax(x)))
        ax.set_ylim((np.amin(y),np.amax(y)))
        ax.legend(loc='best',fontsize=18,framealpha=1)
        # Calculate and present the error
        err_low = val - x[:int(len(y)/2)]\
                         [np.argmin(np.abs(np.array(y[:int(len(y)/2)])-chisq-level))]
        err_high = x[int(len(y)/2):]\
                    [np.argmin(np.abs(np.array(y[int(len(y)/2):])-chisq-level))] - val
        print('Resulting uncertainties: {:.5e} (-) {:.5e} (+) {:.5e}'.
                  format(val,err_low,err_high))
        if savename:
            plt.savefig(savename)
        else:
            plt.show()




#!/usr/bin/env python
"""

Background model has stages
SingleFitter goes through stages and fits each with chi^2, then cstat

MultiFitter fits first one with SingleFitter,
then goes through all the others
by setting the parameter values to those of the previous id
and then fitting each stage


"""
import numpy
from numpy import log10
import json
import logging
import os
import astropy.io.fits as pyfits
import scipy.stats
import scipy.optimize
import datetime
import time


def create_ogip_atable(BackArray, BackSpecFile, SourceSpecFile, outfilename, ilo, ihi):
	"""
	Convert the PCA background model (1D array of counts per channel) into a additive Xspec table model.

	Parameters
	------------
	BackSpecFile: str
		background spectrum file (fits)
	BackArray: numpy array
		background model data
	outfilename: str
		where the atable fits file should be written to
	ilo: int
		lower index of valid data in BackArray
	ihi: int
		upper index of valid data in BackArray
	"""
	# code written by Liu Teng
	# adapted by Johannes Buchner

	bkgspec, hdr = pyfits.getdata(BackSpecFile, header=True, extname='SPECTRUM')
	Nchannel = len(bkgspec)

	hdu0 = pyfits.PrimaryHDU()
	nowstr1 = datetime.datetime.fromtimestamp(time.time()).isoformat()
	nowstr = nowstr1[:nowstr1.rfind('.')]
	hdu0.header['CREATOR'] = "SampSpecFit"
	hdu0.header['DATE'] = nowstr
	hdu0.header['HDUCLASS'] = ('OGIP','format conforms to OGIP standard')
	hdu0.header['HDUDOC']   = ('OGIP/92-009','document defining format')
	hdu0.header['HDUCLAS1'] = ('XSPEC TABLE MODEL','model spectra for XSPEC')
	hdu0.header['HDUVERS1'] = ('1.0.0','version of format')
	hdu0.header['MODLNAME'] = ('pcabkg', 'model name (12 chars max)')
	hdu0.header['MODLUNIT'] = ('ph/cm^2/s', 'model units (12 chars max)')
	hdu0.header['REDSHIFT'] = (False,'whether redshift is to be a parameter')
	hdu0.header['ADDMODEL'] = (True,'whether this is an additive model')
	hdu0.header['LOELIMIT'] = (0, 'model value for energies below those tabulated')
	hdu0.header['HIELIMIT'] = (0, 'model value for energies above those tabulated')
	hdu0.header['EXTEND']   = True
	hdu0.header['SIMPLE']   = True

	if SourceSpecFile:
		source_hdr = pyfits.getheader(SourceSpecFile, extname='SPECTRUM')
		backscal = hdr['BACKSCAL'] / source_hdr['BACKSCAL']
	else:
		print()
		print()
		print("Warning: no source file provided. You need to scale the atable by the source BACKSCAL parameter!")
		backscal = hdr['BACKSCAL']
	dtype1 = [('NAME', 'S12'), ('METHOD', '>i4'), ('INITIAL', '>f4'), ('DELTA', '>f4'), ('MINIMUM', '>f4'), ('BOTTOM', '>f4'), ('TOP', '>f4'), ('MAXIMUM', '>f4'), ('NUMBVALS', '>i4'), ('VALUE', '>f4', (3,))]
	parameters = numpy.array([
		('isSource', 0, 1, -1, 0, 0, 1, 1, 2, numpy.array([0, 1, 0])),
	], dtype=dtype1)
	
	hdu1 = pyfits.BinTableHDU(data=parameters)
	hdu1.header['DATE'] = nowstr
	hdu1.header['EXTNAME'] = 'PARAMETERS'
	hdu1.header['HDUCLASS'] = 'OGIP'
	hdu1.header['HDUCLAS1'] = 'XSPEC TABLE MODEL'
	hdu1.header['HDUCLAS2'] = 'PARAMETERS'
	hdu1.header['HDUVERS1'] = '1.0.0'
	hdu1.header['NINTPARM'] = len(parameters)
	hdu1.header['NADDPARM'] = 0

	dtype2 = [('ENERG_LO', '>f4'), ('ENERG_HI', '>f4')]
	hdu2 = pyfits.BinTableHDU(data=numpy.array([(i, i+1) for i in range(0,Nchannel)], dtype=dtype2))
	hdu2.header['ilow'] = ilo+1
	hdu2.header['ihigh'] = ihi+1
	hdu2.header['DATE'] = nowstr
	hdu2.header['EXTNAME']  = 'ENERGIES'
	hdu2.header['HDUCLASS'] = 'OGIP'
	hdu2.header['HDUCLAS1'] = 'XSPEC TABLE MODEL'
	hdu2.header['HDUCLAS2'] = 'ENERGIES'
	hdu2.header['HDUVERS1'] = '1.0.0'
	hdu2.header['TUNIT1']   = 'keV'
	hdu2.header['TUNIT2']   = 'keV'

	dtype = [('PARAMVAL', '>f4', (len(parameters),)), 
	('INTPSPEC', '>f4', (Nchannel,)),
	]
	# store the model in counts / s / cm^2
	bspec = numpy.pad(BackArray, (ilo, Nchannel-ihi)) / hdr['EXPOSURE'] / hdr['AREASCAL']
	print("background spectrum is scaled by:", hdr['EXPOSURE'], hdr['AREASCAL'])
	# for the source, adjust by the background scaling factor
	sspec = bspec / backscal
	print(BackArray.mean())
	print("source spectrum is up-scaled by:", backscal)
	table = numpy.array([
		((0,), bspec),
		((1,), sspec),
	], dtype=dtype)
	hdu3 = pyfits.BinTableHDU(data=table)
	hdu3.header['DATE'] = nowstr
	hdu3.header['EXTNAME']  = 'SPECTRA'
	hdu3.header['TUNIT1']   = 'none'
	hdu3.header['TUNIT2']   = 'ph/channel/s'
	hdu3.header['TUNIT3']   = 'ph/channel/s'
	hdu3.header['HDUCLAS1'] = 'XSPEC TABLE MODEL'
	hdu3.header['HDUCLAS2'] = 'MODEL SPECTRA'
	hdu3.header['HDUVERS1'] = '1.0.0'

	pyfits.HDUList([hdu0,hdu1,hdu2,hdu3]).writeto(outfilename, overwrite=True)
	print("""

-> Xspec atable of the best-fit background created at: %(modelfilename)s.
Keep in mind that you need to use a diagonal RMF and no ARF for this
contribution. To use it in xspec:

    # load source data
    data 1:1 %(sourcefilename)s
    backgrnd 1 none

    # set your energy limits (before dummyrsp!)
    ignore 1:*
    notice 1:0.5-8.0

    # load background data separately
    data 2:2 %(backfilename)s
    ignore 2:*
    notice 2:%(ilo)d-%(ihi)d

    # set unit RMF for background model
    response 2 none
    dummyrsp 0 %(nchan)d %(nchan)d lin 0 1 2:1
    dummyrsp 0 %(nchan)d %(nchan)d lin 0 1 2:2

    # set background PCA model and its parameters:
    model 2:pcabkg atable{%(modelfilename)s}
    1
    1 -1
    0

    # then set your source model
    # model powerlaworwhatever

"""  % dict(
		sourcefilename=SourceSpecFile, modelfilename=outfilename,
		backfilename=BackSpecFile, nchan=Nchannel, ilo=ilo, ihi=ihi)
	)


def create_spectral_files(fitter, result, src_filename):
	ext = fitter.bkg_filename.split('.')[-1]
	f = pyfits.open(fitter.bkg_filename)
	# replace the COUNTS column with our model prediction
	# we need to change the data type from ints to floats
	origdata = f['SPECTRUM'].data
	i = origdata.dtype.names.index('COUNTS')
	dtype = [('COUNTS', '>f4') if name == 'COUNTS' else (name, origdata.dtype.fields[name][0]) for name in origdata.dtype.names]
	v = f['SPECTRUM'].data['COUNTS'] * 0.
	v[fitter.ilo:fitter.ihi] = result
	lnewdata = []
	for j, (origrow, newv) in enumerate(zip(origdata, v)):
		newrow = list(origrow)
		newrow[i] = newv
		lnewdata.append(tuple(newrow))
	newdata = numpy.array(lnewdata, dtype=dtype)
	hdu = pyfits.BinTableHDU(data=newdata, name=f['SPECTRUM'].name)
	for k, v in list(f['SPECTRUM'].header.items()):
		if k not in hdu.header:
			hdu.header[k] = v
	hdulist = [hdu if e.name == 'SPECTRUM' else e for e in f]
	fout = fitter.bkg_filename[:-len(ext)] + 'bstat.' + ext
	hdus = pyfits.HDUList(hdulist)
	hdus.writeto(fout, overwrite=True)
	hdus.close()
	del f
	
	# update source file
	print('creating src file with updated BACKFILE header ...')
	bext = src_filename.split('.')[-1]
	f = pyfits.open(src_filename)
	for e in f:
		if 'BACKFILE' in e.header:
			e.header['BACKFILE'] = fout
	foutsrc = src_filename[:-len(bext)] + 'bstat.' + bext
	f.writeto(foutsrc, overwrite=True)
	f.close()
	return foutsrc

"""
Find background model.

I analysed the background for many instruments, and stored mean and
 principle components. The data file tells us which instrument we deal with,
 so we load the correct file.
First guess: 
 1) PCA decomposition.
 2) Mean scaled, other components zero
The one with the better cstat is kept.
Then start with 0 components and add 1, 2 components until no improvement
in AIC/cstat.

"""
logf = logging.getLogger('bxa.Fitter')
logf.setLevel(logging.INFO)

class Parameter(object):
	def __init__(self, modelname, name, val, min, max, hard_min=min, hard_max=max):
		self.modelname = modelname
		self.name = name
		self.val = val
		self.min = min
		self.max = max

# Model 
class PCAModel(object):
	def __init__(self, data):
		modelname = 'pca'
		#self.U = data['U']
		self.V = numpy.matrix(data['components'])
		self.mean = data['mean']
		self.s = data['values']
		self.ilo = data['ilo']
		self.ihi = data['ihi']
		
		p0 = Parameter(modelname=modelname, name='lognorm', val=1, min=-5, max=20,
			hard_min=-100, hard_max=100)
		pars = [p0]
		for i in range(len(self.s)):
			pi = Parameter(modelname=modelname, name='PC%d' % (i+1), 
				val=0, min=-20, max=20,
				hard_min=-1e300, hard_max=1e300)
			pars.append(pi)
		self.pars = pars
	
	def calc(self, p):
		try:
			lognorm = p[0]
			pars = numpy.array(p[1:])
			y = numpy.array(pars * self.V.T + self.mean).flatten()
			cts = (10**y - 1) * 10**lognorm
			return cts
		except Exception as e:
			print("Exception in PCA model:", e, p)
			raise e

def gaussmodel_calc(x, LineE, Sigma, norm):
	cts = norm * numpy.exp(-0.5 * ((x - LineE)/Sigma)**2)
	return cts

def minimize(f, x0):
	return scipy.optimize.minimize(f, x0=x0, method='Nelder-Mead', tol=0.001, options=dict(maxiter=1000, disp=False))

class PCAFitter(object):
	def __init__(self, bkg_filename):
		f = pyfits.open(bkg_filename)
		self.bkg_filename = bkg_filename
		self.data = f[1].data['COUNTS'].astype(int)
		self.ndata = len(self.data)
		self.ngaussians = 0
		logf.info('PCAFitter starting')
		hdr = f[1].header
		telescope = hdr.get('TELESCOP','')
		instrument = hdr.get('INSTRUME', '')
		if telescope == '' and instrument == '':
			raise Exception('ERROR: The TELESCOP/INSTRUME headers are not set in the data file.')
		for folder in os.environ.get('BKGMODELDIR', '.'), os.path.dirname(__file__):
			filename = os.path.join(folder, ('%s_%s_%d.json' % (telescope, instrument, self.ndata)).lower())
			if os.path.exists(filename):
				self.load(filename)
				return
			filename = os.path.join(folder, ('%s_%d.json' % (telescope, self.ndata)).lower())
			if os.path.exists(filename):
				self.load(filename)
				return
		raise Exception('ERROR: Could not load PCA components for this detector (%s %s, %d channels). Try the SingleFitter instead.' % (telescope, instrument, self.ndata))
	
	def load(self, filename):
		logf.info('loading PCA information from %s' % (filename))
		data = json.load(open(filename))
		self.pca = dict()
		for k, v in data.items():
			self.pca[k] = numpy.array(v)
		nactivedata = self.pca['ihi'] - self.pca['ilo']
		assert self.pca['hi'].shape == (nactivedata,), 'spectrum has different number of channels: %d vs %s' % (len(self.pca['hi']), self.ndata)
		assert self.pca['lo'].shape == self.pca['hi'].shape
		assert self.pca['mean'].shape == self.pca['hi'].shape
		assert len(self.pca['components']) == nactivedata
		assert nactivedata <= self.ndata
		ilo = int(self.pca['ilo'])
		ihi = int(self.pca['ihi'])
		self.cts = self.data[ilo:ihi]
		self.x = numpy.arange(ihi-ilo)
		self.ilo = ilo
		self.ihi = ihi
	
	def decompose(self):
		#ilo = int(self.pca['ilo'])
		#ihi = int(self.pca['ihi'])
		#lo = self.pca['lo']
		#hi = self.pca['hi']
		mean = self.pca['mean']
		V = numpy.matrix(self.pca['components'])
		s = self.pca['values']
		#U = self.pca['U']
		ncts = self.cts.sum()
		logf.info('have %d background counts for deconvolution' % ncts)
		y = numpy.log10(self.cts * 1. / ncts  + 1.0)
		z = (y - mean) * V
		assert z.shape == (1,len(s)), z.shape
		return numpy.array([numpy.log10(ncts + 0.1)] + z.tolist()[0])
	
	def calc_bkg_stat(self, pred):
		if pred is None:
			return 1e100
		logls = scipy.stats.poisson(pred).logpmf(self.cts)
		stat = numpy.where(numpy.isfinite(logls), -2 * logls, 1e100 * (1 + numpy.max(self.cts) - numpy.nanmax(pred))**2).sum()
		#print("stat: %.1f" % stat)  #  pred, self.cts)
		return stat

	def complete_parameters(self, pars):
		newpars = numpy.zeros(len(self.model.pars))
		for i, p in enumerate(pars):
			newpars[i] = p
		return newpars

	def predict(self, pars):
		# add gaussians
		#ngausspars = 3 * self.ngaussians
		pred = 0
		for i in range(self.ngaussians):
			LineE, logSigma, lognorm = pars[3*i:3*(i+1)]
			if LineE < 0 or LineE > self.ndata:
				#print('LineE out of range:', LineE)
				return None
			# should not be narrower than a single bin, otherwise we get
			# overfitting
			if logSigma < log10(3.0) or logSigma > log10(self.ndata):
				#print('logSigma out of range:', logSigma, 10**logSigma)
				return None
			if lognorm < -6 or lognorm > 10:
				#print('lognorm out of range:', lognorm, 10**lognorm)
				return None
			pred = pred + gaussmodel_calc(self.x, LineE, 10**logSigma, 10**lognorm)
		pcapars = pars[3*self.ngaussians:]
		newpars = self.complete_parameters(pcapars)
		pred = pred + self.model.calc(newpars)
		return pred
	
	def calc_prior(self, pars):
		return 0 # no prior
		ngausspars = 3 * self.ngaussians
		#gausspars = pars[:ngausspars]
		pcapars = numpy.asarray(pars[ngausspars:])
		return numpy.sum((pcapars)**2) # gaussian prior with std 1
		
	
	def calc_bkg_stat_wrapped_gaussians(self, pars):
		return self.calc_bkg_stat(self.predict(pars)) + self.calc_prior(pars)
		
	def fit(self):
		# try a PCA decomposition of this spectrum
		logf.info('fitting background using PCA method')
		initial = self.decompose()
		logf.info('fit: initial PCA decomposition: %s' % (initial))
		# use neldermead first
		bkgmodel = PCAModel(data=self.pca)
		npars = len(bkgmodel.pars)
		self.model = bkgmodel
		for p, v in zip(bkgmodel.pars, initial):
			p.val = v
		
		predictions = []
		logf.info('fit: first fit ...')
		r = minimize(self.calc_bkg_stat_wrapped_gaussians, x0=initial)
		final = r.x
		predictions.append(self.predict(final))
		logf.info('fit: parameters: %s' % (final))
		initial_v = self.calc_bkg_stat_wrapped_gaussians(final)
		logf.info('fit: stat: %s' % (initial_v))
		
		# lets try from zero
		logf.info('fit: second full fit from zero ...')
		zero = numpy.zeros(len(initial))
		zero[0] = log10(numpy.mean(self.data))
		r = minimize(self.calc_bkg_stat_wrapped_gaussians, x0=zero)
		final_v0 = r.x
		initial_v0 = self.calc_bkg_stat_wrapped_gaussians(final)
		predictions.append(self.predict(final_v0))
		logf.info('fit: parameters: %s' % (final))
		logf.info('fit: stat: %s' % (initial_v0))
		
		# pick the better starting point
		if initial_v0 < initial_v:
			logf.info('fit: using zero-fit')
			initial_v = initial_v0
			final = final_v0
		else:
			logf.info('fit: using decomposed-fit')
		
		for p, v in zip(bkgmodel.pars, final):
			p.val = v
		
		# start with the full fit and remove(freeze) parameters
		print('%d parameters, stat=%.2f' % (len(initial), initial_v))
		results = [(2 * len(final) + initial_v, final, len(final), initial_v)]
		for i in range(len(initial)-1, 0, -1):
			initial = final[:i]
			r = minimize(self.calc_bkg_stat_wrapped_gaussians, x0=initial)
			final = self.complete_parameters(r.x)
			predictions.append(self.predict(final))
			v = self.calc_bkg_stat_wrapped_gaussians(final)
			print('--> %d parameters, stat=%.2f' % (i, v))
			results.insert(0, (v + 2*i, final, i, v))
		
		print()
		print('Background PCA fitting AIC results:')
		print('-----------------------------------')
		print()
		print('stat Ncomp AIC')
		for aic, params, nparams, val in results:
			print('%-05.1f %2d %-05.1f' % (val, nparams, aic))
		aic, final, nparams, val = min(results)
		for p, v in zip(bkgmodel.pars, final):
			p.val = v
		
		print()
		print('Increasing parameters again...')
		# now increase the number of parameters again
		#results = [(aic, final, nparams, val)]
		last_aic, last_final, last_nparams, _ = aic, final, nparams, val
		for i in range(last_nparams, npars):
			next_nparams = i + 1
			initial = last_final[:i]
			r = minimize(self.calc_bkg_stat_wrapped_gaussians, x0=initial)
			v = self.calc_bkg_stat_wrapped_gaussians(r.x)
			next_final = self.complete_parameters(r.x)

			next_aic = v + 2*next_nparams
			if next_aic < last_aic:
				# accept
				print('%d parameters, aic=%.2f ** accepting' % (next_nparams, next_aic))
				last_aic, last_final, last_nparams, _ = next_aic, next_final, next_nparams, v
			else:
				print('%d parameters, aic=%.2f' % (next_nparams, next_aic))
			# stop if we are 3 parameters ahead what we needed
			if next_nparams >= last_nparams + 3:
				break
		
		print('Final choice: %d parameters, aic=%.2f' % (last_nparams, last_aic))
		# reset to the last good solution
		for p, v in zip(bkgmodel.pars, last_final):
			p.val = v
		
		last_pred = self.predict(last_final)
		predictions.append(last_pred)
		#if self.cts.sum() < len(self.cts): # do not add gaussians to low-count background spectra
		#	print 'Not adding Gaussians, because low count'
		#	return last_pred, predictions
		#return last_final
		del i
		for gi in range(10):
			print()
			print('Adding Gaussian#%d' % (gi+1))
			# find largest discrepancy
			
			y = self.cts.cumsum()
			z = last_pred.cumsum()
			diff = numpy.abs(y - z)
			xi = numpy.argmax(diff)
			print('largest remaining discrepancy at %d, need %d counts' % (xi, diff[xi]))
			power = diff[xi]
			print('placing gaussian there ...')
			# we work in energy bins, not energy
			initial = [xi, log10(4.), log10(power)] + list(last_final)[:last_nparams]
			self.ngaussians = gi + 1
			print('initial guess:', self.calc_bkg_stat_wrapped_gaussians(initial))
			r = minimize(self.calc_bkg_stat_wrapped_gaussians, x0=initial)
			next_final = r.x
			v = self.calc_bkg_stat_wrapped_gaussians(next_final)
			next_pred = self.predict(next_final)
			predictions.append(next_pred)
			
			next_nparams = last_nparams + 3
			next_aic = v + 2 * next_nparams
			print('with Gaussian:', next_aic, '; change: %.1f (negative is good)' % (next_aic - last_aic))
			if next_aic < last_aic:
				print('accepting')
				last_aic, last_final, last_nparams, _ = next_aic, next_final, next_nparams, v
				last_pred = next_pred
			else:
				print('not significant, rejecting')
				# reset to previous model
				return last_pred, predictions

def main():
	import sys
	#logging.basicConfig(filename='bxa.log',level=logging.DEBUG)
	#logFormatter = logging.Formatter("[%(name)s %(levelname)s]: %(message)s")
	logFormatter = logging.Formatter("%(levelname)s: %(message)s")
	consoleHandler = logging.StreamHandler()
	consoleHandler.setFormatter(logFormatter)
	consoleHandler.setLevel(logging.INFO)
	logging.getLogger().addHandler(consoleHandler)
	logf.setLevel(logging.INFO)

	if len(sys.argv) not in (2, 3):
		print('SYNOPSIS: %s <bkg.pi> [<src.pi>] ' % sys.argv[0])
		sys.exit(1)
	background_file = sys.argv[1]
	source_file = sys.argv[2] if len(sys.argv) > 2 else None
	del sys
	fitter = PCAFitter(background_file)
	result, predictions = fitter.fit()
	data = fitter.cts

	# write out bkg file
	print('creating bstat bkg file ...')

	if background_file.endswith('.fits'):
		numpy.savetxt(background_file[:-5] + '.dat', result)
	with open(background_file + '.bstat.out', 'w') as fout:
		numpy.savetxt(fout, numpy.transpose([data, result]))
	
	# plot fit
	print('plotting...')
	m = max(data.sum(), result.sum())
	x = numpy.arange(fitter.ilo, fitter.ihi)
	import matplotlib.pyplot as plt
	plt.figure()
	plt.title('data-model should follow 1:1 dashed line')
	plt.plot(data.cumsum(), result.cumsum(), '-', color='r', lw=2, alpha=0.7)
	plt.plot([0, m], [0, m] , '--', color='gray', lw=0.4)
	plt.ylabel('Cumulative counts (data)')
	plt.ylabel('Cumulative counts (model)')
	plt.savefig(background_file + '.bstat_cum.pdf', bbox_inches='tight')
	plt.close()
	plt.figure()
	plt.title('model (red) should describe data (black)')
	for r in predictions:
		if r is not None:
			plt.plot(x, r, '-', color='orange', lw=1, alpha=0.3)
	plt.plot(x, result, '-', color='r', lw=2)
	plt.plot(x, data, 'o ', color='k')
	plt.ylim(0, max(result.max(), data.max()))
	plt.xlabel('channel')
	plt.ylabel('counts')
	plt.savefig(background_file + '.bstat.pdf', bbox_inches='tight')
	plt.yscale('log')
	plt.ylim(0.01, max(result.max(), data.max()))
	plt.savefig(background_file + '.bstat_log.pdf', bbox_inches='tight')
	plt.close()
	print()
	print('-> Check that %s is a 1:1 line' % (background_file + '.bstat_cum.pdf'))
	if source_file:
		foutsrc = create_spectral_files(fitter, result, source_file)
		print()
		print('-> In xspec, to load the data with BStat statistic, run:')
		print()
		print('   data %s' % foutsrc)
		print('   statistic pstat # (to use BStat) ')
		print()
		create_ogip_atable(result, background_file, source_file,
			outfilename=background_file + '_model.fits', ilo=fitter.ilo, ihi=fitter.ihi)

if __name__ == '__main__':
	main()

__dir__ = [PCAFitter, PCAModel]



#!/usr/bin/env python
# -*- coding: utf-8 -*- 

"""
Statistics and Plotting for quantile-quantile analysis for model discovery
"""

import numpy
import matplotlib.pyplot as plt
import os
from xspec import Plot


def KSstat(data, model):
	"""
	Kolmogorov-Smirnov statistic: maximum deviance between data and model
	"""
	modelc = model.cumsum() / model.sum()
	datac = data.cumsum() / data.sum()
	ks = numpy.abs(modelc - datac).max()
	return ks

def CvMstat(data, model):
	"""
	Cramér–von Mises statistic: Takes all deviances into account
	"""
	modelc = model.cumsum()
	datac = data.cumsum()
	maxmodelc = modelc.max()
	cvm = ((modelc / maxmodelc - datac / datac.max())**2 * model / maxmodelc).mean()
	return cvm

def ADstat(data, model):
	"""
	Anderson-Darling statistic: Takes all deviances into account
	more weight on tails than CvM.
	"""
	modelc = model.cumsum()
	datac = data.cumsum()
	maxmodelc = modelc.max()
	valid = numpy.logical_and(modelc > 0, maxmodelc - modelc > 0)
	modelc = modelc[valid] / maxmodelc
	datac = datac[valid] / datac.max()
	model = model[valid] / maxmodelc
	assert (modelc > 0).all(), ['ADstat has zero cumulative denominator', modelc]
	assert (maxmodelc - modelc > 0).all(), ['ADstat has zero=1-1 cumulative denominator', maxmodelc - modelc]
	ad = ((modelc - datac)**2 / (modelc * (maxmodelc - modelc)) * model).sum()
	return ad

def qq_plot(bins, data, model, markers = [0.2, 1, 2, 5, 10], unit = '', annotate = True):
	"""
	Create a quantile-quantile plot for model discovery (deviations in data from model).

	* bins: energies/channel
	* data: amount observed
	* model: amount predicted
	* markers: list of energies/channels (whichever the current plotting xaxis unit)
	* unit: unit to append to marker
	* annotate: add information to the plot

	"""

	assert numpy.isfinite(data).all(), data
	assert numpy.isfinite(model).all(), model
	datac = data.cumsum()
	modelc = model.cumsum()
	
	plt.plot(modelc, datac, color='red', drawstyle='steps', linewidth=3)
	
	for m in markers:
		mask = bins >= m
		# first true value
		i = mask.argmax()
		if mask[i]:
			plt.plot(modelc[i], datac[i], marker='+', color='k')
			if datac[i] < modelc[i]:
				textkwargs = dict(va='top', ha='left')
			else:
				textkwargs = dict(va='bottom', ha='right')
			plt.text(modelc[i], datac[i], '%s%s' % (m, unit), **textkwargs)

	u = max(datac[-1], modelc[-1])
	plt.plot([0, u], [0, u], ls='--', color='grey')
	plt.xlim(0, u)
	plt.ylim(0, u)
	plt.title('QQ-plot')
	plt.xlabel('integrated model')
	plt.ylabel('cumulative data counts')

	if annotate:
		plt.text(u/2, u/2, 'data excess', va='bottom', ha='center', rotation=45, color='grey', size=8)
		plt.text(u/2, u/2, 'model excess', va='center', ha='left', rotation=45, color='grey', size=8)
		stats = dict(
			ks = KSstat(data, model),
			cvm = CvMstat(data, model),
			ad = ADstat(data, model),
		)
		
		text = """K-S = %(ks).3f
C-vM = %(cvm).5f
A-D = %(ad)e"""  % stats
		
		plt.text(u*0.98, u*0.02, text, va='bottom', ha='right', color='grey')

def xspecfilenamenormalise(path):
	"""
	Xspec gets a bit confused if there are "." in the filename
	So we replace . with _ in the filename.
	But if we replace it as part of the parent directory, we cannot write
	there, so this only alters the filename, not the entire path.
	"""
	if '/' in path:
		parts = path.split('/')
		parts = (parts[:-1] + [xspecfilenamenormalise(parts[-1])])
		return '/'.join(parts)
	return path.replace('.', '_')

def qq(prefix, markers=5, annotate=True):
	"""
	Create a quantile-quantile plot for model discovery (deviations in data from model).

	The current data and model is used, so call *set_best_fit(analyzer, transformations)*
	before, to get the qq plot at the best fit.

	* markers: list of energies/channels (whichever the current plotting xaxis unit)
	  or number of equally spaced markers between minimum+maximum.
	* annotate: add information to the plot

	"""
	
	olddevice = Plot.device
	Plot.device = '/null'
	tmpfilename = '%swdatatmp.qdp' % xspecfilenamenormalise(prefix)
	if os.path.exists(tmpfilename):
		os.remove(tmpfilename)

	while len(Plot.commands) > 0:
		Plot.delCommand(1)
	Plot.addCommand('wdata "%s"' % tmpfilename.replace('.qdp', ''))
	
	Plot.background = True
	Plot("counts")
	content = numpy.genfromtxt(tmpfilename, skip_header=3)
	os.remove(tmpfilename)
	while len(Plot.commands) > 0:
		Plot.delCommand(1)
	
	if Plot.background:
		bins, width, data, dataerror, background, backgrounderror, model = content[:,0:7].transpose()
	else:
		bins, width, data, dataerror, model = content[:,0:5].transpose()
	
	if not hasattr(markers, '__len__'):
		nmarkers = int(markers)
		decimals = int(-numpy.log10(bins[-1] - bins[0] + 1e-4))
		markers = numpy.linspace(bins[0], bins[-1], nmarkers+2)[1:-1]
		markers = set(numpy.round(markers, decimals=decimals))
	
	# make qq plot, with 1:1 line
	qq_plot(bins=bins, data=data * width * 2, model=model * width * 2, 
		markers = markers, annotate = annotate, unit=Plot.xAxis)
	
	Plot.device = olddevice



#!/usr/bin/env python
import numpy
import sys
import astropy.io.fits as pyfits

changes_needed = []

def should_be(fitsfile, header, keyword, expected=None):
	v = header.get(keyword)
	if expected is None:
		print('%s: %s' % (keyword, v))
		return []
	elif v == expected:
		print('%s: %s (as expected)' % (keyword, v))
		return []
	else:
		print('%s: %s (expected: %s)' % (keyword, v, expected))
		return [(fitsfile, header, keyword, expected)]
	
def apply_changes(f, filename):
	changed_anything = False
	for fitsfile, header, keyword, expected in changes_needed:
		if fitsfile != f: continue
		header[keyword] = expected
		changed_anything = True
	if changed_anything:
		print('   -> writing ',filename) # + '.withkeywords.fits'
		#fitsfile.writeto(filename + '.withkeywords.fits', clobber=True)
		fitsfile.writeto(filename, overwrite=True)
	else:
		print('   -> no changes needed')

if len(sys.argv) != 5:
	print('SYNOPSIS: fixkeywords.py src.pi bkg.pi rmf.rmf arf.arf')
	print() 
	print('Checks and corrects the keywords connecting the files')
	print('Corrects zero energy bounds')
	print() 
	print('Johannes Buchner (C) 2017')
	sys.exit(1)

src = sys.argv[1]
bkg = sys.argv[2]
rmf = sys.argv[3]
arf = sys.argv[4]

fsrc = pyfits.open(src)
fsrch = fsrc[1].header
print('Source file:', src)
print('AREASCAL:', fsrch.get('AREASCAL'))
print('BACKSCAL:', fsrch.get('BACKSCAL'))
print('EXPOSURE:', fsrch.get('EXPOSURE'))
changes_needed += should_be(fsrc, fsrch, 'BACKFILE', bkg)
changes_needed += should_be(fsrc, fsrch, 'RESPFILE', rmf)
changes_needed += should_be(fsrc, fsrch, 'ANCRFILE', arf)
apply_changes(fsrc, src)

print()
changes_needed = []
fbkg = pyfits.open(bkg)
fbkgh = fbkg[1].header
print('Background file:', bkg)
print('AREASCAL:', fbkgh.get('AREASCAL'))
print('BACKSCAL:', fbkgh.get('BACKSCAL'))
print('EXPOSURE:', fbkgh.get('EXPOSURE'))
changes_needed += should_be(fbkg, fbkgh, 'RESPFILE', rmf)
changes_needed += should_be(fbkg, fbkgh, 'ANCRFILE', arf)
apply_changes(fbkg, bkg)

print()
print('Response file:', rmf)
f = pyfits.open(rmf)
SMALL = 0.001
if f['EBOUNDS'].data[0]['E_MIN'] < SMALL or f['MATRIX'].data[0]['ENERG_LO'] < SMALL:
	f['EBOUNDS'].data[0]['E_MIN'] = SMALL
	f['MATRIX'].data[0]['ENERG_LO'] = SMALL
	f.writeto(rmf, overwrite=True)
	print('  -> energy bounds corrected')
else:
	print('  -> file is ok')

print()
print('Ancillary file:', arf)
f = pyfits.open(arf)
if f['SPECRESP'].data[0]['ENERG_LO'] < SMALL:
	f['SPECRESP'].data[0]['ENERG_LO'] = SMALL
	f.writeto(arf, overwrite=True)
	print('  -> energy bounds corrected')
else:
	print('  -> file is ok')



#!/usr/bin/env python
import astropy.io.fits as pyfits
import requests

url = "https://www.swift.ac.uk/analysis/nhtot/donhtot.php"

def parse_response(html_response):
	""" parses the html response and extracts the weighted total NH value """
	akey = "headers='htotw'>"
	a = html_response.index(akey) + len(akey)
	l = html_response[a:].index("</td>")
	part = html_response[a:a+l]
	if part.endswith('</sup>'):
		part = part[:-len('</sup>')]
		base, expo = part.split(' &times;10<sup>')
		return float(base) * 10**float(expo)
	else:
		return float(part)

def get_gal_nh(ra, dec):
	""" asks swift.ac.uk for the total NH value at (ra,dec) """
	r = requests.post(url, data=dict(
		equinox=2000, Coords="%s %s" % (ra, dec),
		submit='Calculate NH')
		)
	nh = parse_response(r.text)
	return nh

if __name__ == '__main__':
	import sys, os
	cache = []
	if len(sys.argv) == 1:
		print("SYNOPSIS: gal.py filename1.pha filename2.pha")
		print()
		print("gal.py requests the total Milky Way column density from swift.ac.uk")
		print("which is then stored in filename1.pha.nh as a plain ASCII number")
		sys.exit(0)
	for infile in sys.argv[1:]:
		outfile = infile + '.nh'
		if not os.path.exists(outfile):
			f = pyfits.open(infile)
			for i in 0, 1:
				header = f[i].header
				print('looking for RA, DEC headers ...')
				if 'RA_OBJ' in header:
					print('using RA_OBJ, DEC_OBJ headers ...')
					ra  = header['RA_OBJ']
					dec = header['DEC_OBJ']
					break
				elif 'RA_TARG' in header:
					print('using RA_TARG, DEC_TARG headers ...')
					ra  = header['RA_TARG']
					dec = header['DEC_TARG']
					break
			nhs = [nhc for rac, decc, nhc in cache if rac == ra and decc == decc]
			if len(nhs) > 0:
				print('same as a previous one')
				nh = nhs[0]
			else:
				print('requesting galactic NH from swift.ac.uk...')
				nh = get_gal_nh(ra, dec)
				cache.append((ra, dec, nh))
				
			print(('writing to %s ...' % outfile))
			open(outfile, 'w').write("%e\n" % nh)
			del ra, dec
		else:
			print(('File %s already exists.' % outfile))



"""
Background models for various telescopes and instruments
"""

import numpy
import json
import logging
import warnings
import os

if 'MAKESPHINXDOC' not in os.environ:
	import sherpa.astro.ui as ui
	from sherpa.stats import Cash, CStat
	from sherpa.models.parameter import Parameter
	from sherpa.models import ArithmeticModel, CompositeModel
	from sherpa.astro.ui import *
	from sherpa.astro.instrument import RSPModelNoPHA, RMFModelNoPHA
else:
	CompositeModel, ArithmeticModel = object, object

class BackgroundModel(object):
	pass

logbm = logging.getLogger('bxa.BackgroundModel')
logbm.setLevel(logging.INFO)

"""
Background model for Chandra, as for the CDFS.

Uses a flat continuum, two broad gaussians and 8 narrow instrumental lines.
"""
class ChandraBackground(BackgroundModel):
	def __init__(self, storage):
		self.storage = storage
		i = self.storage.i
		self.centers = [1.486, 1.739, 2.142, 7.478, 8.012, 8.265, 8.4939, 9.7133]
		continuum, softsoftend, softend = box1d('continuum_%s' % i), gauss1d('softsoftend_%s' % i), gauss1d('softend_%s' % i)
		line1, line2, line3, line4, line5, line6, line7, line8 = [
			gauss1d('line%d_%s' % (j, i)) for j in range(1, 9)]

		self.lines = [line1, line2, line3, line4, line5, line6, line7, line8]
		self.abslines = [line5, line6, line7]
		self.linelines = [line1, line2, line3, line4, line8]
		#lines = linelines + abslines
		self.boxes = [continuum]

		self.softboxes = [softend, softsoftend]
		self.plboxes = self.boxes + self.softboxes
		
		logbm.info('creating Chandra background model')
		for l, c in zip(self.lines, self.centers):
			l.pos = c
			l.pos.min = c - 0.1
			l.pos.max = c + 0.1

		for l in self.lines:
			l.ampl = 2000
			l.ampl.min = 1e-8
			l.ampl.max = 1e12
			l.fwhm = 0.02
			l.fwhm.min = 0.02
			l.pos.freeze()

		for l in self.abslines:
			l.fwhm.max = 0.4

		for l in self.linelines:
			l.fwhm.max = 0.1

		# narrow soft end
		softsoftend.pos = 0.3
		softsoftend.pos.min = 0
		softsoftend.pos.max = 0.6
		softsoftend.pos.freeze()
		softsoftend.fwhm = 0.5
		softsoftend.fwhm.min = 0.2
		softsoftend.fwhm.max = 0.7
		softsoftend.fwhm.freeze()

		# wide gaussian
		softend.pos = 0
		softend.pos.max = 1
		softend.pos.min = -1
		softend.pos.freeze()
		softend.fwhm = 3.8
		softend.fwhm.min = 2
		softend.fwhm.max = 7
		softend.fwhm.freeze()

		for b in self.plboxes:
			b.ampl.val = 1
			b.ampl.freeze()
		changepoints = [0.2, 0.8, 1.3, 2.5, 8.2, 8.4, 8.75, 12]
		changepoints = [0., 12]

		for b,clo,chi in zip(self.boxes, changepoints[:-1], changepoints[1:]):
			b.xlow = clo
			b.xhi = chi
			b.xlow.freeze()
			b.xhi.freeze()
			b.ampl.min = 1e-8
			b.ampl.max = 1e8
			b.ampl.val = 1
		
		contlevel = const1d("contlevel_%s" % i)
		contlevel.integrate = False
		softlevel = const1d("softlevel_%s" % i)
		softlevel.integrate = False
		softsoftlevel = const1d("softsoftlevel_%s" % i)
		softsoftlevel.integrate = False
		contlevel.c0.min = 1e-6
		softlevel.c0.min = 1e-4
		softsoftlevel.c0.min = 1e-4
		contlevel.c0.max = 1e10
		softlevel.c0.max = 1e10
		softsoftlevel.c0.max = 1e10
		
		# init
		contlevel.c0.val = 1e-3
		softlevel.c0.val = 1e1
		softsoftlevel.c0.val = 1e1
		self.stages = ['continuum', 'softfeatures'] + ['line%d' % i for i, line in enumerate(self.linelines + self.abslines)] + ['full']
		self.contlevel = contlevel
		self.softlevel = softlevel
		self.softsoftlevel = softsoftlevel
		
		
	def set_zero(self):
		for l in self.lines + self.plboxes:
			l.ampl.min = 0
			l.ampl.val = 0
	
	"""
	range over which this model is valid
	"""
	def set_filter(self):
		ignore(None, 0.4)
		ignore(9.8, None)
		notice(0.4, 9.8)
	def set_model(self, stage):
		i = self.storage.i
		withsoft = stage not in ('continuum')
		withlines = stage not in ('continuum', 'softfeatures')
		logbm.info('stage "%s" for %s ...' % (stage, i))
		continuum = self.boxes[0]
		contlevel = self.contlevel
		softlevel = self.softlevel
		softsoftlevel = self.softsoftlevel
		softend, softsoftend = self.softboxes
		bg_model = continuum * contlevel
		bunitrsp = self.storage.bunitrsp
		set_bkg_model(i, bg_model)
		set_bkg_full_model(i, bunitrsp(bg_model))
		logbm.debug('zooming to %.1f %.1f' % (2.5, 7))
		ignore(None, 2.5)
		ignore(7, None)
		notice(2.5, 7)
		self.stagepars = [contlevel.c0]
		self.pars = list(self.stagepars)
		if stage == 'continuum': 
			return
		
		logbm.debug('adding soft end...')
		bg_model += (softend * softlevel + softsoftend * softsoftlevel) * contlevel
		delete_bkg_model(i)
		set_bkg_model(i, bg_model)
		set_bkg_full_model(i, bunitrsp(bg_model))
		ignore(None, 0.4)
		ignore(2., None)
		notice(0.4, 2)
		logbm.debug('zooming to %.1f %.1f' % (0.4, 2.))
		self.stagepars = [softsoftlevel.c0, softlevel.c0, softsoftlevel.c0]
		self.pars += self.stagepars
		if stage == 'softfeatures': 
			return
		logbm.debug('adding lines...')
		for j, l in enumerate(self.linelines + self.abslines):
			bg_model += l * contlevel
			set_bkg_model(i, bg_model)
			set_bkg_full_model(i, bunitrsp(bg_model))
			if l.ampl.frozen: continue
			lo = l.pos.val - max(3*l.fwhm.val, 0.2)
			hi = l.pos.val + max(3*l.fwhm.val, 0.2)
			logbm.debug('zooming to %.1f %.1f' % (lo, hi))
			ignore(None, lo)
			ignore(hi, None)
			notice(lo, hi)
			self.stagepars = [l.ampl]
			self.pars += self.stagepars + [l.fwhm]
			if stage == 'line%d' % j: 
				return
			self.set_filter()
		# finally, full fit
		self.stagepars = self.pars

"""
Background model for XMM/PN, as for the XMM-XXL.

Developed by Richard Sturm at MPE
Written for Sherpa by Zhu Liu
Adapted into this framework by Johannes Buchner
"""
class XMMPNBackground(BackgroundModel):
	pncenters = [1.49165, 1.49165, 4.53177, 5.42516, 6.38155, 7.48675, 8.04087, 8.04087, 8.60924, 8.89395, 9.56160]
	pnlinewidth = [5.73813E-02, 3.63469E-02, 6.10487E-02, 7.08380E-02, 9.59053E-02, 6.52422E-02, 9.48594E-02, 6.26174E-05, 0.120893, 0.114254, 0.108717]
	pnlinenorm = [7.81356E-03, 3.96601E-03, 7.30727E-04, 4.96413E-04, 5.31295E-9, 6.84796E-04, 3.01564E-02, 1.41847E-04, 8.87887E-03, 5.75592E-03, 1.71367E-03]
	def __init__(self, storage, galactic_NH = 1e20):
		self.storage = storage
		i = self.storage.i

		self.pnrsp = get_response(i)
		self.pnbrsp = get_response(i,bkg_id=1)
		self.pnscale = get_bkg_scale(i)
		copy_data(i,1000+2)
		self.pnbunitrsp = get_response(1000+2, bkg_id=1)
		delete_data(1000+2)
		self.pnbkgcons = xsconstant('pncons_%s' % i)
		self.pnbkgspline1 = xsspline('pnspline1_%s' % i)
		self.pnbkgexpdec = xsexpdec('pnexpdec_%s' % i)
		self.pnbkgsmedge1 = xssmedge('pnsmedge1_%s' % i)
		self.pnbkgspline2 = xsspline('pnspline2_%s' % i)
		self.pnbkgsmedge2 = xssmedge('pnsmedge2_%s' % i)
		self.pnbkginspl = xspowerlaw('pnbkpl_%s' % i)
		self.pnlines = []
		for j in range(1, 11+1):
			comp = xsgaussian('pngau%d_%s' % (j,i))
			self.pnlines.append(comp)
			setattr(self, 'pnbkgline%d' % j, comp)
		self.pnbkgpl = xspowerlaw('pnpnexpl_%s' % i)
		self.pnbkgapec = xsapec('pnapec_%s' % i)
		self.pnbkglcapec = xsapec('pnlcapec_%s' % i)
		
		self.galabs = xsphabs('absgal_%s' % i)
		self.galabs.nH = galactic_NH / 1e22
		self.galabs.nH.freeze()
		
		self.pnfixwid = [self.pnbkgline2, self.pnbkgline3, self.pnbkgline4, self.pnbkgline5, self.pnbkgline6, self.pnbkgline8, self.pnbkgline11]
		self.pnfree = [self.pnbkgline1, self.pnbkgline7, self.pnbkgline9, self.pnbkgline10]
		
		#define PN background model
		#pn_bkg = self.pnbunitrsp(self.pnbkgcons*(self.pnbkgspline1*self.pnbkgexpdec + self.pnbkgsmedge1*self.pnbkgsmedge2*(self.pnbkgspline2*self.pnbkginspl + 
		#	self.pnbkgline1 + self.pnbkgline2 + self.pnbkgline3 + self.pnbkgline4 + self.pnbkgline5 + self.pnbkgline6 + self.pnbkgline7 + self.pnbkgline8 + 
		#	self.pnbkgline9 + self.pnbkgline10 + self.pnbkgline11))) + self.pnbrsp(self.galabs*(self.pnbkgpl+self.pnbkgapec)+self.pnbkglcapec)
		
		# Define source model: galabs*(intgalabs*(srcpl+torus+pnsoftpl))
		#set_model(i,galabs*(intgalabs*(xspowerlaw.srcpl)))
		#set_full_model(i,(pnrsp(galabs*(intgalabs*(xspowerlaw.srcpl))) + pnscale*(pn_bkg)))
		
		for l, c in zip(self.pnlines, self.pncenters):
			l.LineE = c
			l.LineE.min = c - 0.05
			l.LineE.max = c + 0.05
		
		self.pnbkgline2.LineE = self.pnbkgline1.LineE
		self.pnbkgline8.LineE = self.pnbkgline7.LineE
		
		for l, s in zip(self.pnlines, self.pnlinewidth):
			l.Sigma = s

		for l in self.pnfree:
			l.Sigma.min = 1E-5
			l.Sigma.max = 0.2
		
		for l in self.pnfixwid:
			l.Sigma.freeze()
		
		for l, n in zip(self.pnlines, self.pnlinenorm):
			l.norm = n
			l.norm.min = 1E-10
			l.norm.max = 1E10
			
		# Scaling constant
		self.pnbkgcons.factor = 1.0
		self.pnbkgcons.factor.freeze()
		
		# thermal radiation
		self.pnbkgapec.kT = 0.286928
		self.pnbkgapec.kT.min = 0.008
		self.pnbkgapec.kT.max = 64
		self.pnbkgapec.Abundanc = 1.0
		self.pnbkgapec.Abundanc.freeze()
		self.pnbkgapec.Redshift = 0.0
		self.pnbkgapec.Redshift.freeze()
		self.pnbkgapec.norm = 5.58410E-05
		self.pnbkgapec.norm.min = 1e-10
		self.pnbkgapec.norm.max = 1e10
		
		# local thermal radiation
		self.pnbkglcapec.kT = 0.1
		self.pnbkglcapec.kT.freeze()
		self.pnbkglcapec.Abundanc = 1.0
		self.pnbkglcapec.Abundanc.freeze()
		self.pnbkglcapec.Redshift = 0.0
		self.pnbkglcapec.Redshift.freeze()
		self.pnbkglcapec.norm = 3.89164E-05
		self.pnbkglcapec.norm.min = 1e-10
		self.pnbkglcapec.norm.max = 1e10
		
		# exponential decay
		self.pnbkgexpdec.factor = 44.3418
		self.pnbkgexpdec.factor.min = 0
		self.pnbkgexpdec.factor.max = 100
		self.pnbkgexpdec.norm = 6830.89
		self.pnbkgexpdec.norm.freeze()
		
		# Smear function
		self.pnbkgsmedge1.edgeE = 0.538408
		self.pnbkgsmedge1.edgeE.freeze()
		self.pnbkgsmedge1.MaxTau = 1.40238
		self.pnbkgsmedge1.MaxTau.min = 0 
		self.pnbkgsmedge1.MaxTau.max = 10
		self.pnbkgsmedge1.index = -2.67000
		self.pnbkgsmedge1.index.freeze()
		self.pnbkgsmedge1.width = 0.313365
		self.pnbkgsmedge1.width.min = 0.01
		self.pnbkgsmedge1.width.max = 100
		
		self.pnbkgsmedge2.edgeE = 1.38826
		self.pnbkgsmedge2.edgeE.freeze()
		self.pnbkgsmedge2.MaxTau.min = 0 
		self.pnbkgsmedge2.MaxTau.max = 10
		self.pnbkgsmedge2.MaxTau = 9.37167
		self.pnbkgsmedge2.index = -2.67000
		self.pnbkgsmedge2.index.freeze()
		self.pnbkgsmedge2.width = 5.7642
		self.pnbkgsmedge2.width.min = 0.01
		self.pnbkgsmedge2.width.max = 100
		
		# Spline funtion
		self.pnbkgspline1.Estart = 0.200000
		self.pnbkgspline1.Estart.freeze()
		self.pnbkgspline1.Ystart = -1.31506
		self.pnbkgspline1.Ystart.min = -1e+6
		self.pnbkgspline1.Ystart.max = 1e+6
		self.pnbkgspline1.Yend = 1064.16
		self.pnbkgspline1.Yend.min = -1e+6
		self.pnbkgspline1.Yend.max = 1e+6
		self.pnbkgspline1.YPstart = -106.183
		self.pnbkgspline1.YPstart.min = -1e+6
		self.pnbkgspline1.YPstart.max = 1e+6
		self.pnbkgspline1.YPend = -366.092
		self.pnbkgspline1.YPend.min = -1e+6
		self.pnbkgspline1.YPend.max = 1e6
		self.pnbkgspline1.Eend = 1.74715
		self.pnbkgspline1.Eend.min = 0
		self.pnbkgspline1.Eend.max = 100
		
		self.pnbkgspline2.Estart = 3.29056
		self.pnbkgspline2.Ystart = 1.00643
		self.pnbkgspline2.Ystart.min = -1e+6
		self.pnbkgspline2.Ystart.max = 1e+6
		self.pnbkgspline2.Yend = 0.887026
		self.pnbkgspline2.Yend.min = -1e+6
		self.pnbkgspline2.Yend.max = 1e+6
		self.pnbkgspline2.YPstart = -0.278401
		self.pnbkgspline2.YPstart.min = -1e+6
		self.pnbkgspline2.YPstart.max = 1e+6
		self.pnbkgspline2.YPend = 4.84809E-03
		self.pnbkgspline2.YPend.min = -1e+6
		self.pnbkgspline2.YPend.max = 1e6
		self.pnbkgspline2.Eend = 7.32701
		self.pnbkgspline2.Eend.min = 0
		self.pnbkgspline2.Eend.max = 100
		
		# Background power law
		self.pnbkginspl.PhoIndex = 0.279
		self.pnbkginspl.PhoIndex.min = -2
		self.pnbkginspl.PhoIndex.max = 9
		self.pnbkginspl.norm = 8.23614E-03
		self.pnbkginspl.norm.min = 1E-10
		self.pnbkginspl.norm.max = 1E6
		
		# Extragalactic background
		self.pnbkgpl.PhoIndex = 1.46
		self.pnbkgpl.PhoIndex.freeze()
		self.pnbkgpl.norm = 1.25288E-05
		self.pnbkgpl.norm.min = 1e-10
		self.pnbkgpl.norm.max = 1e3
		
		self.stages = []
		self.stage_models = {}
		self.stage_models_full = {}
		
		base = self.pnbkgspline2*self.pnbkginspl
		logbm.info( '    XMMPNBackground: creating stages for %s' % i)
		self.addStage('1', base)
		self.addStage('2', base + self.pnbkgline2)
		self.addStage('3', base + self.pnbkgline1 + self.pnbkgline2)
		self.addStage('4', base + self.pnbkgline1 + self.pnbkgline2 + self.pnbkgline7 + self.pnbkgline8)
		self.addStage('5', base + self.pnbkgline1 + self.pnbkgline2 + self.pnbkgline7 + self.pnbkgline8 + self.pnbkgline9 + self.pnbkgline10)
		lines = self.pnbkgline1 + self.pnbkgline2 + self.pnbkgline3 + self.pnbkgline4 + self.pnbkgline5 + self.pnbkgline6 + self.pnbkgline7 + self.pnbkgline8 + self.pnbkgline9 + self.pnbkgline10 + self.pnbkgline11
		self.addStage('6', base + lines)
		self.addStage('7', base + lines, self.galabs*(self.pnbkgapec))
		self.addStage('8', base + lines, self.galabs*(self.pnbkgapec+self.pnbkgpl))
		self.addStage('9', base + lines, self.galabs*(self.pnbkgapec+self.pnbkgpl)+self.pnbkglcapec)
		logbm.info( '    XMMPNBackground: setup for %s completed' % i)

	def addStage(self, name, instrument_bg, external_bg=None):
		model = self.pnbkgcons*(self.pnbkgspline1*self.pnbkgexpdec + self.pnbkgsmedge1*self.pnbkgsmedge2*(instrument_bg))
		full_model = self.pnbunitrsp(self.pnbkgcons*(self.pnbkgspline1*self.pnbkgexpdec + self.pnbkgsmedge1*self.pnbkgsmedge2*(instrument_bg)))
		if external_bg is not None:
			model = model + external_bg
			full_model = full_model + self.pnrsp(external_bg)
		self.stage_models[name] = model
		self.stage_models_full[name] = full_model
		self.stages.append(name)
		
	"""
	range over which this model is valid
	"""
	def set_filter(self):
		i = self.storage.i
		notice(None, None)
		ignore(None, 0.3)
		ignore(10.0, None)
	
	def set_model(self, stage):
		i = self.storage.i
		
		logbm.info( '    XMMPNBackground: going to stage %s on %s' % (stage, i))
		bg_model = self.stage_models[stage]
		bg_model_full = self.stage_models_full[stage]
		self.stagepars = list(bg_model.pars)
		self.pars = list(self.stagepars)
		
		delete_bkg_model(i)
		logbm.info( '    XMMPNBackground: setting model')
		set_bkg_model(i, bg_model)
		set_bkg_full_model(i, bg_model_full)
		self.set_filter()
"""
Background model for Swift/XRT.

Uses a flat continuum, two broad gaussians and 8 narrow instrumental lines.
"""
class SwiftXRTBackground(BackgroundModel):
	def __init__(self, storage):
		self.storage = storage
		
		i = self.storage.i
		logbm.info( '    SwiftXRTBackground: setting up for %s' % i)
		dip = box1d('dip_%s' % i)
		pbknpl = xsbknpower('pbknpl_%s' % i)
		g1, g2, g3, g4 = [gauss1d('gauss%d_%s' % (j,i)) for j in [1, 2, 3, 4]]

		pbknpl.BreakE.min = 0.2
		pbknpl.BreakE.max = 5
		pbknpl.BreakE.val = 2
		pbknpl.PhoIndx1.max = 4
		pbknpl.PhoIndx2.max = 4
		pbknpl.PhoIndx1.min = 0.8
		pbknpl.PhoIndx2.min = 0.8
		pbknpl.PhoIndx1.val = 2
		pbknpl.PhoIndx2.val = 1.5
		pbknpl.norm.min = 1e-10
		pbknpl.norm.max = 1
		pbknpl.norm.val = 0.004
		
		lines = [(0.1, 0.7, 1.1), (2, 2.15, 2.5), (1, 1.2, 1.4), (0, 0.4, 0.5)]
		

		for g, (lo, mid, hi) in zip([g1, g2, g3, g4], lines):
			g.pos.val = mid
			g.pos.min = lo
			g.pos.max = hi
			g.fwhm.val = 0.2
			g.fwhm.max = 1
			g.fwhm.min = 0.01
			g.ampl.val = 0.01
			g.ampl.max = 1
			g.ampl.min = 1e-6

		dip.xlow = 2
		dip.xhi = 3
		dip.xlow.min = 1.75
		dip.xlow.max = 2.25
		dip.xhi.min = 2.75
		dip.xhi.max = 3.25
		dip.ampl.val = 0.5
		dip.ampl.min = 1e-3
		dip.ampl.max = 1 - 1e-3
		self._pars = [pbknpl, dip, g1, g2, g3, g4]
		self.stages = [2, 3, 4, 5, 6, 7]
	def set_filter(self):
		logbm.debug('SwiftXRTBackground: setting filter 0.3-5keV')
		ignore(None, 0.3)
		ignore(5, None)
		notice(0.3, 5)

	def set_model(self, stage):
		i = self.storage.i
		[pbknpl, dip, g1, g2, g3, g4] = self._pars
		
		logbm.info('stage "%s" for ID=%s ...' % (stage, i))
		bunitrsp = self.storage.bunitrsp
		model = stage
		self.bg_model = pbknpl
		self.stagepars = list(pbknpl.pars)
		if model > 1:
			self.bg_model = self.bg_model + g1 
			self.stagepars += list(g1.pars)
		if model > 2:
			self.bg_model = self.bg_model + g2
			self.stagepars = list(g2.pars)
		if model > 3:
			self.bg_model = self.bg_model + g3
			self.stagepars = list(g3.pars)
		if model > 4:
			self.bg_model = (1 - dip) * self.bg_model
			self.stagepars = list(dip.pars)
		if model > 5:
			self.bg_model = self.bg_model + g4
			self.stagepars = list(g4.pars)
		self.pars = [p for p in self.bg_model.pars if not p.link]
		if model > 6: # finally, full fit
			self.stagepars = self.pars
		set_bkg_model(i, self.bg_model)
		set_bkg_full_model(i, bunitrsp(self.bg_model))
		self.set_filter()
		logbm.debug('background model set for stage "%s" for ID=%s' % (stage, i))

class SwiftXRTWTBackground(SwiftXRTBackground):
	def set_filter(self):
		logbm.debug('SwiftXRTBackground: setting filter 0.4-5keV')
		ignore(None, 0.4)
		ignore(5, None)
		notice(0.4, 5)

__all__ = ['SwiftXRTBackground', 'SwiftXRTWTBackground', 'ChandraBackground']



"""
Copyright 2021 Johannes Buchner

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

import numpy as np
import sys, os
import astropy.io.fits as pyfits
import subprocess

def remove(filename):
	if os.path.exists(filename): 
		os.unlink(filename)

def run(*args, **kwargs):
	print("    running command:", *args)
	subprocess.check_call(*args, **kwargs)

def sum_pha(outfile, filenames, backscals, areascals, rel_weights, **kwargs):
	print()
	print("creating '%s' ..." % outfile)
	remove(outfile)
	run(['mathpha', 'expr=' + '+'.join(filenames), "outfil=" + outfile, 
		'units=C', 'exposure=CALC', 'properr=NO', 'errmeth=POISS-0', 'areascal=NULL',
		'ncomments=1', "comment1=Created_by_addspec.py_alpha", 'chatter=5'])

	# update AREASCAL, BACKSCAL keywords:
	areascal = (areascals * rel_weights).sum()
	backscal = (backscals * rel_weights).sum()
	print("    EXPOSURE and counts were summed")
	print("    relative weights for scale factors:", rel_weights)
	print("    averaged AREASCAL:", areascal)
	print("    averaged BACKSCAL:", backscal)
	with open('.tmp.modhead', 'w') as fout:
		fout.write("AREASCAL %.20f\n" % areascal)
		fout.write("BACKSCAL %.20f\n" % backscal)
		for k, v in kwargs.items():
			fout.write("%s %s\n" % (k, v))
	
	run(['fmodhead', outfile + '[SPECTRUM]', '.tmp.modhead'])
	os.unlink('.tmp.modhead')



def main(outprefix, filenames):

	N = len(filenames)
	print("files:", N, filenames)

	backscals = np.empty(N)
	areascals = np.empty(N)
	exposures = np.empty(N)
	bbackscals = np.empty(N)
	bareascals = np.empty(N)
	bexposures = np.empty(N)
	arfs = []
	rmfs = []
	barfs = []
	brmfs = []

	bfilenames = []

	for i, filename in enumerate(filenames):
		header = pyfits.getheader(filename, extname='SPECTRUM')
		backscals[i] = header['BACKSCAL']
		areascals[i] = header['AREASCAL']
		exposures[i] = header['EXPOSURE']
		arf = header['ANCRFILE']
		rmf = header['RESPFILE']
		arfs.append(arf)
		rmfs.append(rmf)

		backfile = header['BACKFILE']
		bfilenames.append(backfile)
		bheader = pyfits.getheader(backfile, extname='SPECTRUM')
		bbackscals[i] = bheader['BACKSCAL']
		bareascals[i] = bheader['AREASCAL']
		bexposures[i] = bheader['EXPOSURE']
		barf = bheader['ANCRFILE']
		brmf = bheader['RESPFILE']
		barfs.append(barf if str(barf).lower() != 'none' else arf)
		brmfs.append(brmf if str(brmf).lower() != 'none' else rmf)
		
		del arf, rmf, barf, brmf, backfile

	print("Files:", filenames, bfilenames)


	print("BACKSCAL:", backscals, bbackscals)
	print("AREASCAL:", areascals, bareascals)
	print("EXPOSURE:", exposures, bexposures)

	weights = areascals * exposures
	rel_weights = weights / weights.sum()
	weightstr = ' '.join(['%e' % w for w in rel_weights])
	bweights = bareascals * bexposures
	rel_bweights = bweights / bweights.sum()
	bweightstr = ' '.join(['%e' % w for w in rel_bweights])
	assert len(arfs) == N
	assert len(rmfs) == N
	assert len(barfs) == N
	assert len(brmfs) == N

	print("combining ARFs:", arfs)
	arf = outprefix + '.arf'
	run(['addarf', ' '.join(arfs), weightstr, arf, 'clobber=yes'])
	barf = outprefix + '_bkg.arf'
	run(['addarf', ' '.join(barfs), bweightstr, barf, 'clobber=yes'])
	print("combining RMFs:", rmfs)
	rmf = outprefix + '.rmf'
	run(['addrmf', ' '.join(rmfs), weightstr, rmf, 'clobber=yes'])
	brmf = outprefix + '_bkg.rmf'
	run(['addrmf', ' '.join(brmfs), bweightstr, brmf, 'clobber=yes'])
	
	outfile = "%s.pha" % outprefix
	boutfile = "%s_bkg.pha" % outprefix

	sum_pha(
		outfile = outfile,
		filenames = filenames,
		areascals = areascals,
		backscals = backscals,
		rel_weights = rel_weights,
		ANCRFILE = arf,
		RESPFILE = rmf,
		BACKFILE = boutfile,
	)
	sum_pha(
		outfile = boutfile,
		filenames = bfilenames,
		areascals = bareascals,
		backscals = bbackscals,
		rel_weights = rel_bweights,
		ANCRFILE = barf,
		RESPFILE = brmf,
	)

if __name__ == '__main__':
	if len(sys.argv) < 3:
		sys.stderr.write("""SYNOPSIS: addspec.py <outprefix> file1.pha file2.pha ...
or
SYNOPSIS: addspec.py <outprefix> @filelist.txt

In the second case, each line of filelist.txt contains a file name.

If as a outprefix "sum" is given, the output files are named
sum.pha sum.arf sum.rmf sum_bkg.pha sum_bkg.arf sum_bkg.rmf

Johannes Buchner (C) 2021, MIT Licence
""")
		sys.exit(1)

	filenames = sys.argv[2:]
	if len(filenames) == 1 and filenames[0][0] == '@':
		filenames = [filename.rstrip() for filename in open(filenames[0][1:])]
	main(
		outprefix = sys.argv[1],
		filenames = filenames,
	)



#!/usr/bin/env python
# -*- coding: utf-8 -*- 
"""
BXA (Bayesian X-ray Analysis) for Sherpa

Copyright: Johannes Buchner (C) 2013-2025
"""

import os
if 'MAKESPHINXDOC' not in os.environ:
	import sherpa.astro.ui as ui
	from sherpa.stats import Cash, CStat

from .cachedmodel import CachedModel

def auto_galactic_absorption(id=None):
	#model = ui._session.get_model(id).model
	filename = ui._session.get_data(id).name + '.nh'
	print(('loading nH from %s (expecting something like 1e21 in there)' % filename))
	nH = float(open(filename).read().strip())
	galabso = ui.xstbabs('galabso%s' % id)
	#galabsmodel = model * galabso
	galabso.nH = nH / 1e22
	print(('setting galactic nH to %s [units of 1e22/cm²]' % (galabso.nH.val)))
	galabso.nH.freeze()
	return galabso
	cgalabso = CachedModel(galabso)
	print('returning cached galaxy absorption model')
	return cgalabso
	#ui._session.set_model(id, galabsmodel)



#!/usr/bin/env python
# -*- coding: utf-8 -*- 

"""
BXA (Bayesian X-ray Analysis) for Xspec

Copyright: Johannes Buchner (C) 2013-2025

Priors
"""

from math import log10

# priors
def create_uniform_prior_for(model, par):
	"""
	Use for location variables (position)
	The uniform prior gives equal weight in non-logarithmic scale.
	"""
	pval, pdelta, pmin, pbottom, ptop, pmax = par.values
	print('  uniform prior for %s between %f and %f ' % (par.name, pbottom, ptop))
	low = float(pbottom)
	spread = float(ptop - pbottom)
	if pbottom > 0 and ptop / pbottom > 100:
		print('   note: this parameter spans several dex. Should it be log-uniform (create_jeffreys_prior_for)?')
	if pmin != pbottom or ptop != pmax:
		raise UserWarning("in BXA v5, priors now use bottom/top soft parameter limits instead of hard parameter limits.")
	def uniform_transform(x): return x * spread + low
	return dict(model=model, index=par._Parameter__index, name=par.name, 
		transform=uniform_transform, aftertransform=lambda x: x)


def create_jeffreys_prior_for(model, par):
	"""deprecated, use create_loguniform_prior_for instead. """
	return create_loguniform_prior_for(model, par)


def create_loguniform_prior_for(model, par):
	"""
	Use for scale variables (order of magnitude)
	The Jeffreys prior gives equal weight to each order of magnitude between the
	minimum and maximum value. Flat in logarithmic scale
	"""
	pval, pdelta, pmin, pbottom, ptop, pmax = par.values
	print('  jeffreys prior for %s between %e and %e ' % (par.name, pbottom, ptop))
	if pbottom == 0:
		raise Exception('You forgot to set reasonable parameter limits on %s' % par.name)
	if pmin != pbottom or ptop != pmax:
		raise UserWarning("in BXA v5, priors now use bottom/top soft parameter limits instead of hard parameter limits.")
	low = log10(pbottom)
	spread = log10(ptop) - log10(pbottom)
	if spread > 10:
		print('   note: this parameter spans *many* dex. Double-check the limits are reasonable.')
	def log_transform(x): return x * spread + low
	def log_after_transform(x): return 10**x
	return dict(model=model, index=par._Parameter__index, name='log(%s)' % par.name, 
		transform=log_transform, aftertransform=log_after_transform)


def create_gaussian_prior_for(model, par, mean, std):
	"""
	Use for informed variables.
	The Gaussian prior weights by a Gaussian in the parameter.
	"""
	import scipy.stats
	pval, pdelta, pmin, pbottom, ptop, pmax = par.values
	if pmin != pbottom or ptop != pmax:
		raise UserWarning("in BXA v5, priors now use bottom/top soft parameter limits instead of hard parameter limits.")
	rv = scipy.stats.norm(mean, std)
	def gauss_transform(x): 
		return max(pbottom, min(ptop, rv.ppf(x)))
	print('  gaussian prior for %s of %f +- %f' % (par.name, mean, std))
	return dict(model=model, index=par._Parameter__index, name=par.name, 
		transform=gauss_transform, aftertransform=lambda x: x)


def create_custom_prior_for(model, par, transform, aftertransform = lambda x: x):
	"""
	Pass your own prior weighting transformation
	"""
	print('  custom prior for %s' % (par.name))
	return dict(model=model, index=par._Parameter__index, name=par.name, 
		transform=transform, aftertransform=aftertransform)


def create_prior_function(transformations):
	"""
	Creates a single prior transformation function from parameter transformations
	"""

	def prior(cube, ndim, nparams):
		for i, t in enumerate(transformations):
			transform = t['transform']
			cube[i] = transform(cube[i])

	return prior



#!/usr/bin/env python
"""

Kernel density estimation-based background model

It requires no learned models, the continuous shape is learned from the data.
Therefore relatively well-sampled backgrounds are needed.

Bins with zero counts are okay. Rebinning is not required.

"""
import numpy
from numpy import log10
import logging
import astropy.io.fits as pyfits
import scipy.stats
import scipy.optimize
import datetime
import time


def create_ogip_atable(BackArray, BackSpecFile, SourceSpecFile, outfilename, ilo, ihi):
	"""
	Convert the background model (1D array of counts per channel) into a additive Xspec table model.

	Parameters
	------------
	BackSpecFile: str
		background spectrum file (fits)
	BackArray: numpy array
		background model data
	outfilename: str
		where the atable fits file should be written to
	ilo: int
		lower index of valid data in BackArray
	ihi: int
		upper index of valid data in BackArray
	"""
	# code written by Liu Teng
	# adapted by Johannes Buchner

	bkgspec, hdr = pyfits.getdata(BackSpecFile, header=True, extname='SPECTRUM')
	Nchannel = len(bkgspec)

	hdu0 = pyfits.PrimaryHDU()
	nowstr1 = datetime.datetime.fromtimestamp(time.time()).isoformat()
	nowstr = nowstr1[:nowstr1.rfind('.')]
	hdu0.header['CREATOR'] = "SampSpecFit"
	hdu0.header['DATE'] = nowstr
	hdu0.header['HDUCLASS'] = ('OGIP','format conforms to OGIP standard')
	hdu0.header['HDUDOC']   = ('OGIP/92-009','document defining format')
	hdu0.header['HDUCLAS1'] = ('XSPEC TABLE MODEL','model spectra for XSPEC')
	hdu0.header['HDUVERS1'] = ('1.0.0','version of format')
	hdu0.header['MODLNAME'] = ('kdebkg', 'model name (12 chars max)')
	hdu0.header['MODLUNIT'] = ('ph/cm^2/s', 'model units (12 chars max)')
	hdu0.header['REDSHIFT'] = (False,'whether redshift is to be a parameter')
	hdu0.header['ADDMODEL'] = (True,'whether this is an additive model')
	hdu0.header['LOELIMIT'] = (0, 'model value for energies below those tabulated')
	hdu0.header['HIELIMIT'] = (0, 'model value for energies above those tabulated')
	hdu0.header['EXTEND']   = True
	hdu0.header['SIMPLE']   = True

	if SourceSpecFile:
		source_hdr = pyfits.getheader(SourceSpecFile, extname='SPECTRUM')
		backscal = hdr['BACKSCAL'] / source_hdr['BACKSCAL']
	else:
		print()
		print()
		print("Warning: no source file provided. You need to scale the atable by the source BACKSCAL parameter!")
		backscal = hdr['BACKSCAL']
	dtype1 = [('NAME', 'S12'), ('METHOD', '>i4'), ('INITIAL', '>f4'), ('DELTA', '>f4'), ('MINIMUM', '>f4'), ('BOTTOM', '>f4'), ('TOP', '>f4'), ('MAXIMUM', '>f4'), ('NUMBVALS', '>i4'), ('VALUE', '>f4', (3,))]
	parameters = numpy.array([
		('isSource', 0, 1, -1, 0, 0, 1, 1, 2, numpy.array([0, 1, 0])),
	], dtype=dtype1)
	
	hdu1 = pyfits.BinTableHDU(data=parameters)
	hdu1.header['DATE'] = nowstr
	hdu1.header['EXTNAME'] = 'PARAMETERS'
	hdu1.header['HDUCLASS'] = 'OGIP'
	hdu1.header['HDUCLAS1'] = 'XSPEC TABLE MODEL'
	hdu1.header['HDUCLAS2'] = 'PARAMETERS'
	hdu1.header['HDUVERS1'] = '1.0.0'
	hdu1.header['NINTPARM'] = len(parameters)
	hdu1.header['NADDPARM'] = 0

	dtype2 = [('ENERG_LO', '>f4'), ('ENERG_HI', '>f4')]
	hdu2 = pyfits.BinTableHDU(data=numpy.array([(i, i+1) for i in range(0,Nchannel)], dtype=dtype2))
	hdu2.header['ilow'] = ilo+1
	hdu2.header['ihigh'] = ihi+1
	hdu2.header['DATE'] = nowstr
	hdu2.header['EXTNAME']  = 'ENERGIES'
	hdu2.header['HDUCLASS'] = 'OGIP'
	hdu2.header['HDUCLAS1'] = 'XSPEC TABLE MODEL'
	hdu2.header['HDUCLAS2'] = 'ENERGIES'
	hdu2.header['HDUVERS1'] = '1.0.0'
	hdu2.header['TUNIT1']   = 'keV'
	hdu2.header['TUNIT2']   = 'keV'

	dtype = [('PARAMVAL', '>f4', (len(parameters),)), 
	('INTPSPEC', '>f4', (Nchannel,)),
	]
	# store the model in counts / s / cm^2
	bspec = numpy.pad(BackArray, (ilo, Nchannel-ihi)) / hdr['EXPOSURE'] / hdr['AREASCAL']
	print("background spectrum is scaled by:", hdr['EXPOSURE'], hdr['AREASCAL'])
	# for the source, adjust by the background scaling factor
	sspec = bspec / backscal
	print(BackArray.mean())
	print("source spectrum is up-scaled by:", backscal)
	table = numpy.array([
		((0,), bspec),
		((1,), sspec),
	], dtype=dtype)
	hdu3 = pyfits.BinTableHDU(data=table)
	hdu3.header['DATE'] = nowstr
	hdu3.header['EXTNAME']  = 'SPECTRA'
	hdu3.header['TUNIT1']   = 'none'
	hdu3.header['TUNIT2']   = 'ph/channel/s'
	hdu3.header['TUNIT3']   = 'ph/channel/s'
	hdu3.header['HDUCLAS1'] = 'XSPEC TABLE MODEL'
	hdu3.header['HDUCLAS2'] = 'MODEL SPECTRA'
	hdu3.header['HDUVERS1'] = '1.0.0'

	pyfits.HDUList([hdu0,hdu1,hdu2,hdu3]).writeto(outfilename, overwrite=True)
	print("""

-> Xspec atable of the best-fit background created at: %(modelfilename)s.
Keep in mind that you need to use a diagonal RMF and no ARF for this
contribution. To use it in xspec:

    # load source data
    data 1:1 %(sourcefilename)s
    backgrnd 1 none

    # set your energy limits (before dummyrsp!)
    ignore 1:*
    notice 1:0.5-8.0

    # load background data separately
    data 2:2 %(backfilename)s
    ignore 2:*
    notice 2:%(ilo)d-%(ihi)d

    # set unit RMF for background model
    response 2 none
    dummyrsp 0 %(nchan)d %(nchan)d lin 0 1 2:1
    dummyrsp 0 %(nchan)d %(nchan)d lin 0 1 2:2

    # set background PCA model and its parameters:
    model 2:pcabkg atable{%(modelfilename)s}
    1
    1 -1
    0

    # then set your source model
    # model powerlaworwhatever

"""  % dict(
		sourcefilename=SourceSpecFile, modelfilename=outfilename,
		backfilename=BackSpecFile, nchan=Nchannel, ilo=ilo, ihi=ihi)
	)


def create_spectral_files(fitter, result, src_filename):
	ext = fitter.bkg_filename.split('.')[-1]
	f = pyfits.open(fitter.bkg_filename)
	# replace the COUNTS column with our model prediction
	# we need to change the data type from ints to floats
	origdata = f['SPECTRUM'].data
	i = origdata.dtype.names.index('COUNTS')
	dtype = [('COUNTS', '>f4') if name == 'COUNTS' else (name, origdata.dtype.fields[name][0]) for name in origdata.dtype.names]
	v = f['SPECTRUM'].data['COUNTS'] * 0.
	v[fitter.ilo:fitter.ihi] = result
	lnewdata = []
	for j, (origrow, newv) in enumerate(zip(origdata, v)):
		newrow = list(origrow)
		newrow[i] = newv
		lnewdata.append(tuple(newrow))
	newdata = numpy.array(lnewdata, dtype=dtype)
	hdu = pyfits.BinTableHDU(data=newdata, name=f['SPECTRUM'].name)
	for k, v in list(f['SPECTRUM'].header.items()):
		if k not in hdu.header:
			hdu.header[k] = v
	hdulist = [hdu if e.name == 'SPECTRUM' else e for e in f]
	fout = fitter.bkg_filename[:-len(ext)] + 'bstat.' + ext
	hdus = pyfits.HDUList(hdulist)
	hdus.writeto(fout, overwrite=True)
	hdus.close()
	del f
	
	# update source file
	print('creating src file with updated BACKFILE header ...')
	bext = src_filename.split('.')[-1]
	f = pyfits.open(src_filename)
	for e in f:
		if 'BACKFILE' in e.header:
			e.header['BACKFILE'] = fout
	foutsrc = src_filename[:-len(bext)] + 'bstat.' + bext
	f.writeto(foutsrc, overwrite=True)
	f.close()
	return foutsrc

class KDEFitter(object):
	""" a KDE approximation of the background spectrum """
	def __init__(self, bkg_filename):
		f = pyfits.open(bkg_filename)
		self.bkg_filename = bkg_filename
		self.cts = f[1].data['COUNTS'].astype(int)
		self.ndata = len(self.cts)
		self.ngaussians = 0
		self.ilo = 0
		self.ihi = self.ndata

	def fit(self, bw_method=None):
		# get photon counts as samples of the channels
		samples = []
		for chan, counts in enumerate(self.cts):
			samples += [chan] * counts

		# apply KDE
		kde = scipy.stats.gaussian_kde(samples, bw_method=bw_method)
		density = kde(numpy.arange(self.ndata))

		# make a model that matches the observed counts
		model = density / density.sum() * self.cts.sum()
		return model, [model]



"""
Find background model.

I analysed the background for many instruments, and stored mean and
 principle components. The data file tells us which instrument we deal with,
 so we load the correct file.
First guess: 
 1) PCA decomposition.
 2) Mean scaled, other components zero
The one with the better cstat is kept.
Then start with 0 components and add 1, 2 components until no improvement
in AIC/cstat.

"""
logf = logging.getLogger('bxa.Fitter')
logf.setLevel(logging.INFO)

def main():
	import sys
	logFormatter = logging.Formatter("%(levelname)s: %(message)s")
	consoleHandler = logging.StreamHandler()
	consoleHandler.setFormatter(logFormatter)
	consoleHandler.setLevel(logging.INFO)
	logging.getLogger().addHandler(consoleHandler)
	logf.setLevel(logging.INFO)

	if len(sys.argv) not in (2, 3):
		print('SYNOPSIS: %s <bkg.pi> [<src.pi>] ' % sys.argv[0])
		sys.exit(1)
	background_file = sys.argv[1]
	source_file = sys.argv[2] if len(sys.argv) > 2 else None
	del sys
	fitter = KDEFitter(background_file)
	result, predictions = fitter.fit()
	data = fitter.cts

	# write out bkg file
	print('creating bstat bkg file ...')

	if background_file.endswith('.fits'):
		numpy.savetxt(background_file[:-5] + '.dat', result)
	with open(background_file + '.bstat.out', 'w') as fout:
		numpy.savetxt(fout, numpy.transpose([data, result]))
	
	# plot fit
	print('plotting...')
	m = max(data.sum(), result.sum())
	x = numpy.arange(fitter.ilo, fitter.ihi)
	import matplotlib.pyplot as plt
	plt.figure()
	plt.title('data-model should follow 1:1 dashed line')
	plt.plot(data.cumsum(), result.cumsum(), '-', color='r', lw=2, alpha=0.7)
	plt.plot([0, m], [0, m] , '--', color='gray', lw=0.4)
	plt.ylabel('Cumulative counts (data)')
	plt.ylabel('Cumulative counts (model)')
	plt.savefig(background_file + '.bstat_cum.pdf', bbox_inches='tight')
	plt.close()
	plt.figure()
	plt.title('model (red) should describe data (black)')
	for r in predictions:
		if r is not None:
			plt.plot(x, r, '-', color='orange', lw=1, alpha=0.3)
	plt.plot(x, result, '-', color='r', lw=2)
	plt.plot(x, data, 'o ', color='k')
	plt.ylim(0, max(result.max(), data.max()))
	plt.xlabel('channel')
	plt.ylabel('counts')
	plt.savefig(background_file + '.bstat.pdf', bbox_inches='tight')
	plt.yscale('log')
	plt.ylim(0.01, max(result.max(), data.max()))
	plt.savefig(background_file + '.bstat_log.pdf', bbox_inches='tight')
	plt.close()
	print()
	print('-> Check that %s is a 1:1 line' % (background_file + '.bstat_cum.pdf'))
	if source_file:
		foutsrc = create_spectral_files(fitter, result, source_file)
		print()
		print('-> In xspec, to load the data with BStat statistic, run:')
		print()
		print('   data %s' % foutsrc)
		print('   statistic pstat # (to use BStat) ')
		print()
		create_ogip_atable(result, background_file, source_file,
			outfilename=background_file + '_model.fits', ilo=fitter.ilo, ihi=fitter.ihi)

if __name__ == '__main__':
	main()

__dir__ = [KDEFitter]



from __future__ import print_function
"""
Lets try something simpler.

Background model has stages
SingleFitter goes through stages and fits each with chi^2, then cstat

MultiFitter fits first one with SingleFitter,
then goes through all the others
by setting the parameter values to those of the previous id
and then fitting each stage


"""

import os
if 'MAKESPHINXDOC' not in os.environ:
	import sherpa.astro.ui as ui
	from sherpa.stats import Cash, CStat
	from sherpa.models.parameter import Parameter
	from sherpa.models import ArithmeticModel, CompositeModel
	from sherpa.astro.ui import *
	from sherpa.astro.instrument import RSPModelNoPHA, RMFModelNoPHA
else:
	CompositeModel, ArithmeticModel = object, object
	load_user_stat = lambda _1, _2, _3: None
import numpy
import json
import logging
import warnings

"""
custom background statistic, similar to chi^2

Sherpa itself does not expose the statistic of the background fit
as far as I can tell. This has the benefit of being rather stable
and giving large, sensitive values when the fit is poor. (CStat tends to 
flatten out)

i: spectrum ID
"""

logi = logging.getLogger('bxa.internal')
logi.setLevel(logging.WARN)

def my_bkg_stat(i):
	p = get_bkg_fit_plot(i)
	m = p.modelplot
	d = p.dataplot
	chi = (((d.y - m.y) / (m.y + 1e-20)) ** 2)
	return chi.mean()

def group_adapt(data, model, nmin = 20):
	i = 0
	while i < len(data):
		for j in range(i, len(data)+1):
			mask = numpy.arange(i, j)
			n = data[mask].sum()
			if n >= nmin or j + 1 >= len(data):
				yield (n, model[mask].sum())
				break
		i = j + 1

def wstatfunc(data, model, staterror=None, syserror=None, weight=None):
	n = data.astype('int').sum()
	l = len(data)
	if n >= l * 20: # no rebinning needed
		r = (((data - model) / model)**2).sum()
		return r, r
	else:
		r = numpy.sum([((d - m)/m)**2 for d, m in group_adapt(data.astype('int'), model, nmin=20)])
		for p in get_bkg_model().pars:
			logi.debug('%s = %.f' % (p.fullname, p.val))
		logi.info('bxaroughstat: %.3f' % r)
		return r, r
def wstatfunc(data, model, staterror=None, syserror=None, weight=None):
	n = data.astype('int').sum()
	l = len(data)
	r = (((data - model) / model)**2).sum()
	return r, r
def custom_staterr_func(data):
	return data**0.5
load_user_stat("bxaroughstat", wstatfunc, custom_staterr_func)


"""
custom routine to find the best normalisation in a robust way.

Often the starting point can be way off and the statistic will not 
give good indications which way to go. So use my_bkg_stat. This routine
tries to go both up and down from the starting point to escape this flatness.

i: spectrum ID
params: list of params to deal with, in sequence
"""
def robust_opt(i, params):
	logi.debug('        robust_opt: my optimization     --- of %s' % i)
	logi.debug('        robust_opt: \tparameters:', ', '.join([p.fullname for p in params]))
	#plot_bkg_fit_resid(i)
	for p in params:
		logi.debug('        robust_opt: \toptimizing %s normalization [%e .. %e]' % (p.fullname, p.min, p.max))
		startval = p.val
		beststat = my_bkg_stat(i)
		bestval = p.val
		logi.info('        robust_opt: \t\tstart val = %e: %e' % (p.val, beststat))
		go_up = True
		go_down = True
		for n in list(3.**numpy.arange(1, 20)) + [None] + list(1.1**numpy.arange(1, 13)):
			if n is None:
				startval = bestval
				logi.info('        robust_opt: \t\trefining from %e' % (startval))
				go_up = True
				go_down = True
				continue
			if go_up and startval * n > p.max:
				logi.info('        robust_opt: \t\thit upper border (%e * %e > %e)' % (startval, n, p.max))
				go_up = False
			if go_down and startval / n < p.min:
				logi.info('\t\thit lower border (%e / %e > %e)' % (startval, n, p.min))
				go_down = False
			if go_up:
				logi.debug('        robust_opt: \t\ttrying %e ^' % (startval * n))
				p.val = startval * n
				newstat = my_bkg_stat(i)
				logi.debug('        robust_opt: \t\tval = %e: %e' % (p.val, newstat))
				if newstat <= beststat:
					bestval = p.val
					beststat = newstat
					logi.debug('        robust_opt: \t\t\timprovement: %e' % p.val)
				if newstat > beststat + 100:
					go_up = False
			if go_down:
				logi.debug('        robust_opt: \t\ttrying %e v' % (startval / n))
				p.val = startval / n
				newstat = my_bkg_stat(i)
				logi.debug('        robust_opt: \t\tval = %e: %e' % (p.val, newstat))
				if newstat + 1e-3 < beststat:
					bestval = p.val
					beststat = newstat
					logi.debug('        robust_opt: \t\t\timprovement: %e' % p.val)
				if newstat > beststat + 100:
					go_down = False
		p.val = bestval
		logi.debug('        robust_opt: \tnew normalization of %s: %e' % (p.fullname, p.val))
		#plot_bkg_fit_resid(i)
	logi.info('        robust_opt: optimization of %s in %s done, reached %.3f' % (', '.join([p.fullname for p in params]), i, beststat))
	if beststat > 0.2:
		logi.info('        robust_opt: no good fit found: %.3f' % beststat)
		if False:
			plot_bkg_fit_resid(i)
			print('press ENTER to continue >> ', end=' ')
			sys.stdin.readline()
	
	logi.debug('        robust_opt: my optimization end ---')

logbs = logging.getLogger('bxa.BackgroundStorage')
logbs.setLevel(logging.WARN)

"""
Loads background file and assigns unit response matrix.

Capabilities for loading and storing previously fitted background models.
"""
class BackgroundStorage(object):
	def __init__(self, backgroundfilename, i, load=False):
		self.backgroundfilename = backgroundfilename
		self.backgroundparamsfile = backgroundfilename + '_bkgparams.json'
		logbs.info('      BackgroundStorage: for background model params of ID=%s, will use "%s" as storage' % (i, self.backgroundparamsfile))
		if load:
			load_pha(i, self.backgroundfilename)
		self.i = i
		self.load_rsp()
	def load_rsp(self):
		i = self.i
		copy_data(i,1000+2)
		unit_arf = get_bkg_arf(1000+2)
		unit_arf.specresp = 0.*unit_arf.specresp + 1.0 # * unit_arf.specresp.mean()
		self.bunitrsp = get_response(1000+2,bkg_id=1)
		delete_data(1000+2)
	def load_bkg_model(self):
		i = self.i
		m = get_bkg_model(i)
		bkg_model = json.load(open(self.backgroundparamsfile, 'r'))
		bkg_pars = bkg_model['params']
		for p in m.pars:
			logbs.info("loaded parameter: \tbkg[%s] = %e" % (p.fullname, bkg_pars[p.fullname]))
			if bkg_pars[p.fullname] > p.max or bkg_pars[p.fullname] < p.min:
				warnings.warn('WARNING: ignoring stored parameter value for %s (val=%s, min=%s, max=%s)' % (p.fullname, p.val, p.min, p.max))
				continue
			p.val = bkg_pars[p.fullname]
		return bkg_model
	
	def store_bkg_model(self, stats, **kwargs):
		i = self.i
		m = get_bkg_model(i)
		params = dict([(p.fullname, p.val) for p in m.pars])
		oldstats = 1e300

		try:
			prev_model = json.load(open(self.backgroundparamsfile, 'r'))
			oldstats = prev_model['stats']
		except IOError as e:
			warnings.warn('store_bkg_model: could not check previously stored model, perhaps this is the first. (Error was: %s)' % e)
		if stats > oldstats + 0.02:
			logbs.warn('store_bkg_model: ERROR: refusing to store worse model! %.3f (new) vs %.3f (old)' % (stats, oldstats))
		else:
			set_analysis(i, "energy", "counts")
			p = get_bkg_fit_plot(i)
			with open(self.backgroundparamsfile, 'w') as f:
				json.dump(dict(params=params, stats=stats,
					plot=dict(x=p.dataplot.x.tolist(), ydata=p.dataplot.y.tolist(), ymodel=p.modelplot.y.tolist()),
					**kwargs), 
					f, indent=4)
			logbs.info('store_bkg_model: stored as "%s"' % self.backgroundparamsfile)
		names = [p.fullname for p in m.pars]
		values = [p.val for p in m.pars]
		return names, values
"""
Robust fitting strategies for background models
"""


logf = logging.getLogger('bxa.Fitter')
logf.setLevel(logging.INFO)

class SingleFitter(object):
	def __init__(self, id, filename, backgroundmodel, load=False):
		""" 
		id: which data id to fit
		
		filename: prefix for where to store background information
		
		backgroundmodel: A background model, such as ChandraBackground
		
		load: whether the background file should be loaded now
		"""
		self.id = id
		logf.info('SingleFitter(for ID=%s, storing to "%s")' % (id, filename))
		logf.debug('  creating backgroundstorage ...')
		b = BackgroundStorage(filename, id, load=load)
		logf.debug('  creating Background Model ...')
		self.bm = backgroundmodel(b)
	def store(self):
		i = self.id
		# reset filter for equal comparison
		self.bm.set_filter()
		newstats = my_bkg_stat(i)
		self.bm.storage.store_bkg_model(newstats, stage=self.stage)
	def tryload(self):
		for stage in self.bm.stages:
			self.prepare_stage(stage=stage)
			props = self.bm.storage.load_bkg_model()
			logf.info('Background loaded, stage %s%s' % (stage, '(last)' if self.bm.stages[-1] == stage else '(more to go)'))
			if stage == props['stage']:
				break
		
	def fit(self, reinit=True, plot=False):
		for stage in self.bm.stages:
			self.prepare_stage(stage=stage)
			self.fit_stage(stage=stage, plot=plot)
		logf.info('Background fit complete.')
	
	#def fit_continue(self, reinit=True, plot=False):
	#	stages = list(self.bm.stages)
	#	for stage in self.bm.stages:
	#		self.prepare_stage(stage=stage)
	#		try:
	#			props = self.bm.storage.load_bkg_model()
	#			logf.info('Background loaded, stage %s%s' % (stage, '(last)' if self.bm.stages[-1] == stage else '(more to go)'))
	#			stages.pop(0)
	#			if stage == props['stage']:
	#				break
	#	for stage in stages:
	#		self.prepare_stage(stage=stage)
	#		self.fit_stage(stage=stage, plot=plot)
			
	
	def prepare_stage(self, stage, prev=None, link=False):
		i = self.id
		logf.info('prepare_stage %s of ID=%s' % (stage, i))
		self.bm.set_model(stage=stage)
		if prev is not None and link:
			otherpars = prev
			this = self.bm
			assert len(otherpars) == len(self.bm.pars), (other, 'incompatible with', self.bm.pars)
			for pa, pb in zip(otherpars, self.bm.pars):
				aname = pa.fullname.split('_')[0]
				bname = pb.fullname.split('_')[0]
				assert aname == bname, ('names should be the same', pa.fullname, pb.fullname)
				logf.debug('   linking %s <- %s' % (pb.fullname, pa.fullname))
				pb = pa
		elif prev is not None:
			for pa, pb in zip(self.bm.pars, prev):
				if pa.link:
					logf.debug('   unlinking %s' % (pa.fullname))
					pa.unlink()
				if pa in self.bm.stagepars:
					pa.val = pb.val
		# only thaw those parameters that are new
		for pa in self.bm.pars:
			if pa in self.bm.stagepars:
				if pa.link: continue
				pa.thaw()
			else:
				pa.freeze()
		logf.info('prepare_stage %s of ID=%s done' % (stage, i))
				
	def fit_stage(self, stage, plot=False):
		i = self.id
		#set_method_opt('ftol', 0.1)
		def doplot():
			if plot:
				s = my_bkg_stat(i)
				logf.debug('fit_stage %s of ID=%s (stat: %.3f)... plotting ... ' % (stage, i, s))
				plot_bkg_fit_resid(i)
				logf.debug('fit_stage %s of ID=%s (stat: %.3f)... plotting done' % (stage, i, s))
		logf.info('fit_stage %s of ID=%s' % (stage, i))
		logf.debug(get_bkg_model(i))
		
		if get_bkg(i).counts.sum() > 500:
			logf.info('fit_stage %s of ID=%s. rough fit ...' % (stage, i))
			prev_filter = get_filter(i)
			logf.debug('grouping')
			group_counts(i, 20)
			set_stat('chi2gehrels')
			logf.debug('plotting')
			doplot()
			
			normparams = [p for p in self.bm.stagepars if p.name in ['ampl', 'norm']]
			if normparams:
				logf.debug('simple optimization')
				robust_opt(i, normparams)
			logf.debug('calling fit_bkg()')
			fit_bkg(i)
			
			logf.debug('plotting')
			doplot()
			ungrouping_bug = False
			logf.debug('ungrouping')
			try:
				ungroup(i)
			except Exception as e:
				logf.warn('ungrouping failed')
				ungrouping_bug = True
			
			logf.debug('setting filters')
			ignore()
			notice(prev_filter)
			self.stage = stage
			logf.debug('storing')
			self.store()
		
		logf.info('fit_stage %s of ID=%s.  fine fit ...' % (stage, i))
		set_method_opt('ftol', 0.001)
		set_stat('cstat')
		doplot()
		fit_bkg(i)
		doplot()
		logf.info('fit_stage %s of ID=%s.  fitted' % (stage, i))
		self.stage = stage
		self.store()
		logf.info('fit_stage %s of ID=%s.  stage done' % (stage, i))
		try:
			ungroup(i)
		except Exception as e:
			logf.warn('ungrouping failed')
		
logmf = logging.getLogger('bxa.MultiFitter')
logmf.setLevel(logging.INFO)

class MultiFitter(object):
	def __init__(self, filename, backgroundmodel, load=True):
		"""
		filename should be a text file, containing all the 
		file names (if load=True) or storage prefixes
		
		backgroundmodel: background model to use (e.g. ChandraBackground)
		
		load: whether the background file should be loaded now
		"""
		ids = [l.strip() for l in open(filename).readlines()]
		names = list(ids)
		#if load:
		#	ids = list(range(1, 1+len(names)))
		logmf.debug('MultiFitter: loading...')
		self.fitters = {}
		for i, name in zip(ids, names):
			self.fitters[i] = SingleFitter(i, name, backgroundmodel, load=load)
		self.ids = sorted(ids, key=lambda i: get_bkg(i).counts.sum(), reverse=True)
		logmf.debug('MultiFitter: loading done')
		
	"""
	MultiFitter fits first one with SingleFitter,
	then goes through all the others
	by setting the parameter values to those of the previous id
	and then fitting each stage
	"""
	def fit(self, **kwargs):
		batchsize = 10
		logmf.info('MultiFitter: splitting work into batches of size %d' % batchsize)
		for part in range(int(numpy.ceil(len(self.ids[1:]) * 1. / batchsize))):
			batchids = self.ids[part*batchsize:part*batchsize+batchsize]
			logmf.debug('MultiFitter: batch with %s' % str(batchids))
			first = batchids[0]
			firstfit = self.fitters[first]
			logmf.info('MultiFitter: fit first ID=%s ...' % first)
			stages = firstfit.bm.stages
			pars_at_stages = {}
			for stage in firstfit.bm.stages:
				firstfit.prepare_stage(stage=stage)
				firstfit.fit_stage(stage=stage, **kwargs)
				pars_at_stages[stage] = list(firstfit.bm.pars)
			logmf.info('MultiFitter: fit first ID=%s done' % first)
			
			# we want to improve our statistics by fitting jointly
			for stage in stages:
				logmf.info('MultiFitter: joint fitting, stage "%s" ...' % stage)
				firstfit.prepare_stage(stage=stage)
				for i in batchids[1:]:
					self.fitters[i].prepare_stage(stage=stage, prev=pars_at_stages[stage], link=True)
				
				logmf.debug('MultiFitter: joint fitting, stage "%s", rough chi^2...' % stage)
				set_method_opt('ftol', 0.1)
				set_stat('chi2gehrels')
				logmf.debug('MultiFitter: joint fitting, stage "%s", rough chi^2 fit...' % stage)
				fit_bkg(*batchids)
				logmf.debug('MultiFitter: joint fitting, stage "%s", rough chi^2 storing...' % stage)
				for i in batchids:
					self.fitters[i].store()
				set_stat('cstat')
				logmf.debug('MultiFitter: joint fitting, stage "%s", cstat...' % stage)
				fit_bkg(*batchids)
				for i in batchids:
					self.fitters[i].store()
				#self.fitters[i].fit_stage(stage=stage, **kwargs)
				pars_at_stages[stage] = self.fitters[i].bm.pars
				logmf.info('MultiFitter: joint fitting, stage "%s" done' % stage)
			# now we release the links and try to improve the fits
			# individually
			for stage in stages:
				logmf.info('MultiFitter: individual fitting, stage "%s" ...' % stage)
				firstfit.prepare_stage(stage=stage)
				logmf.info('MultiFitter: individual fitting, stage "%s" unlinking ' % stage)
				for i in batchids[1:]:
					self.fitters[i].prepare_stage(stage=stage, prev=pars_at_stages[stage], link=False)
				logmf.info('MultiFitter: individual fitting, stage "%s" fitting ' % stage)
				for i in batchids:
					self.fitters[i].fit_stage(stage=stage, **kwargs)
					#pars_at_stages[stage] = self.fitters[i].bm.pars
				logmf.info('MultiFitter: individual fitting, stage "%s" done' % stage)
	def fit_jointly_stage(self, stage, ids, plot=False):
		def doplot(i):
			if plot:
				s = my_bkg_stat(i)
				logmf.debug('fit_jointly_stage %s of ID=%s (stat: %.3f)... plotting ... ' % (stage, i, s))
				plot_bkg_fit_resid(i)
				logmf.debug('fit_jointly_stage %s of ID=%s (stat: %.3f)... plotting done' % (stage, i, s))
		params = []
		prev_filters = []
		for i in ids:
			prev_filters.append(get_filter(i))
			group_counts(i, 20)
			set_stat('chi2gehrels')
			doplot(i)
			params += [p for p in self.bm.stagepars if not p.link and not p.frozen]
		fit_bkg(id=ids[0], otherids=ids[1:])
		doplot(ids[0])
		for i, prev_filter in zip(ids, prev_filters):
			ungroup(i)
			ignore()
			notice(prev_filter)
			self.fitters[i].store()
		logmf.debug('fit_jointly_stage %s: fine fit ...' % (stage))
		
		set_method_opt('ftol', 0.001)
		set_stat('cstat')
		doplot(ids[0])
		fit_bkg(id=ids[0], otherids=ids[1:])
		doplot(ids[0])
		logmf.debug('fit_jointly_stage %s: fitted' % (stage))
		for i in ids:
			self.fitters[i].store()
		logmf.debug('fit_jointly_stage %s: stage done' % (stage))

__all__ = ['SingleFitter', 'MultiFitter', 'BackgroundStorage', 'robust_opt', 'my_bkg_stat']



#!/usr/bin/env python3
"""
Creates background PCA model from a large sample of background spectra
"""
import astropy.io.fits as pyfits
import numpy
import sys

if __name__ == '__main__':
	filenames = sys.argv[1:]
	if len(filenames) == 0:
		sys.stderr.write("""SYNOPSIS: %s <filenames>

Builds a machine-learned background model from a large set of
background spectra (filenames).

Johannes Buchner (C) 2017-2019
""" % sys.argv[0])
		sys.exit(1)
	if len(filenames) == 1:
            filenames = list(map(str.rstrip, open(filenames[0]).readlines()))
	for filename in filenames:
		f = pyfits.open(filename)
		s = f['SPECTRUM']
		print(s.data['COUNTS'].sum(), filename)
		f.close()




#!/usr/bin/env python
# -*- coding: utf-8 -*- 

"""
BXA (Bayesian X-ray Analysis) for Xspec

Copyright: Johannes Buchner (C) 2013-2025
"""

from . import qq
from .sinning import binning

from xspec import Plot
import matplotlib.pyplot as plt
from .priors import *
from .solver import BXASolver, XSilence, create_prior_function


def nested_run(
	transformations, prior_function=None,
	n_live_points=400, evidence_tolerance=0.5,
	outputfiles_basename='chains/', verbose=True, **kwargs
):
	"""
	deprecated. Use BXASolver instead.
	"""
	solver = BXASolver(
		transformations=transformations, prior_function=prior_function, 
		outputfiles_basename=outputfiles_basename)
	return solver.run(n_live_points=n_live_points, evidence_tolerance=evidence_tolerance, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
BXA (Bayesian X-ray Analysis) for Sherpa

Copyright: Johannes Buchner (C) 2013-2025
"""
import os
from math import log10, isnan, isinf
if 'MAKESPHINXDOC' not in os.environ:
	import sherpa.astro.ui as ui
	from sherpa.stats import Cash, CStat

import numpy
from .priors import *
from .galabs import auto_galactic_absorption
from .solver import BXASolver, default_logging


def nested_run(
	id=None, otherids=(), prior=None, parameters=None,
	sampling_efficiency=0.3, evidence_tolerance=0.5,
	n_live_points=400, outputfiles_basename='chains/', **kwargs
):
	"""
	deprecated, use BXASolver instead.
	"""
	solver = BXASolver(
		id=id, otherids=otherids, prior=prior, parameters=parameters,
		outputfiles_basename=outputfiles_basename)
	return solver.run(
		evidence_tolerance=evidence_tolerance,
		n_live_points=n_live_points, **kwargs)



#!/usr/bin/env python
import h5py
import numpy
import sys
from . import plotcolors
import matplotlib.pyplot as plt

def pca(M):
	mean = M.mean(axis=0)
	Moffset = M - mean.reshape((1,-1))
	U, s, Vt = numpy.linalg.svd(Moffset, full_matrices=False)
	V = Vt.T
	print('variance explained:', s**2/len(M))
	c = numpy.cumsum(s**2/len(M))
	c = c / c[-1]
	for cut in 0.80, 0.90, 0.95, 0.99:
		idx, = numpy.where(c>cut)
		n = idx.min() + 1
		print('  --> need %d components to explain %.2f%%' % (n, cut))
	return U, s, V, mean

def pca_predict(U, s, V, mean):
	S = numpy.diag(s)
	return numpy.dot(U, numpy.dot(S, V.T)) + mean.reshape((1,-1))

def pca_get_vectors(s, V, mean):
	#U = numpy.eye(len(s))
	#return pca_predict(U, s, V, mean)
	Sroot = numpy.diag(s**0.5)
	return numpy.dot(Sroot, V.T)

def pca_cut(U, s, V, mean, ncomponents=20):
	return U[:, :ncomponents], s[:ncomponents], V[:,:ncomponents], mean
#def pca_cut_adapt(U, s, V, mean):
#	return U[:, :ncomponents], s[:ncomponents], V[:,:ncomponents], mean

def pca_check(M, U, s, V, mean):
	# if we use only the first 20 PCs the reconstruction is less accurate
	Mhat2 = pca_predict(U, s, V, mean)
	print("Using %d PCs, MSE = %.6G"  % (len(s), numpy.mean((M - Mhat2)**2)))
	return M - Mhat2

def run(cmd, filename, mincts = 50000):
	f = h5py.File(filename, 'r')
	attrs = f['spectra'].attrs
	data = f['spectra'][()]

	if cmd == 'create':
		lo = data.min(axis=0)
		hi = data.max(axis=0)
		i, = numpy.where(lo != hi)
		ilo = i.min() + 1
		ihi = i.max() - 1
		selected = data.sum(axis=1)>mincts
		cts = data[selected,ilo:ihi]
		ncts = cts.sum(axis=1).reshape((-1,1))
		print('channel range:', ilo, ihi, cts.shape)
		print('lowest number of counts: %d' % ncts.min())
		y = numpy.log10(cts * 1. / ncts  + 1.0)

		lo = y.min(axis=0)
		hi = y.max(axis=0)
		#data = (y - lo.reshape((1,-1)) ) / (hi - lo).reshape((1,-1))
		data = y
		print('running pca', data.shape)
		U, s, V, mean = pca(data)
		U, s, V, mean = pca_cut(U, s, V, mean, 5)

		with h5py.File(filename + 'pca.hdf5', 'w') as f:
			f.attrs['ilo'] = ilo
			f.attrs['ihi'] = ihi
			for k, v in attrs.items():
				print(('   storing attribute %s = %s' % (k, v)))
				f.attrs[k] = v
			f.create_dataset("mean", data=mean, compression="gzip", compression_opts=9, shuffle=True)
			f.create_dataset("components", data=V, compression="gzip", compression_opts=9, shuffle=True)
			f.create_dataset("values", data=s, compression="gzip", compression_opts=9, shuffle=True)
			f.create_dataset("U", data=U, compression="gzip", compression_opts=9, shuffle=True)
			f.create_dataset("lo", data=lo, compression="gzip", compression_opts=9, shuffle=True)
			f.create_dataset("hi", data=hi, compression="gzip", compression_opts=9, shuffle=True)
		return filename + 'pca.hdf5'
		
	elif cmd == 'components':
		with h5py.File(filename + 'pca.hdf5', 'r') as f:
			ilo = f.attrs['ilo']
			ihi = f.attrs['ihi']
			mean = f['mean'][()]
			V = f['components'][()]
			s = f['values'][()]
			U = f['U'][()]
			scales = numpy.abs(U).mean(axis=0)
			vectors = pca_get_vectors(s, V, mean)[:6]
			plt.figure(figsize=(12, 4))
			name = '%(INSTRUMENT)s on %(TELESCOPE)s' % f.attrs
			plt.text(0.95, 0.95, name, va='top', ha='right', transform=plt.gca().transAxes, fontsize=20)
			colors = plotcolors.create(len(vectors), cmap='viridis')
			for i, (row, factor, color) in list(enumerate(zip(vectors, scales, colors))):
				if -row.min() > row.max():
					sgn = -1
				else:
					sgn = +1
				sgn = sgn * 3.00
				#sgn = sgn * 2.0**-i
				# make it so that deviation between mean and component is a factor of 10 somewhere
				#delta = numpy.max(numpy.abs(mean + sgn * factor * row))
				#sgn = sgn * 10. / delta
				print(i, row.min(), row.max(), sgn)
				plt.plot(list(range(ilo, ihi)), (10**(mean + sgn * factor * row) - 1) * 2**-i, '-', 
					color=color, alpha=0.75, lw=12 / (i+4), label='PC%s' % (i+1))
				plt.plot(list(range(ilo, ihi)), (10**mean - 1) * 2**-i, '-', color='k', lw=2, label='mean')
			#plt.xscale('log')
			plt.yscale('log')
			yvals = 10**mean - 1
			plt.xlim(ilo, ihi)
			plt.ylim(yvals.min()/40, yvals.max()*2)
			plt.ylabel('Counts [arb. norm.]')
			plt.xlabel('Energy Channel')
			plt.legend(loc='lower right', prop=dict(size=8), ncol=3)
			plt.savefig(filename + 'pca2.pdf', bbox_inches='tight')
			plt.close()
	elif cmd == 'showcomp':		
		with h5py.File(filename + 'pca.hdf5', 'r') as f:
			ilo = f.attrs['ilo']
			ihi = f.attrs['ihi']
			mean = f['mean'][()]
			V = f['components'][()]
			s = f['values'][()]
			U = f['U'][()]
			
			print(mean)
			for i, row in enumerate(pca_get_vectors(s, V, mean)):
				print(row)
				plt.title('PCA component %d' % (i+1))
				plt.plot(mean, '-', color='k')
				plt.plot(mean + row, '-', alpha=0.5)
				plt.show()
	elif cmd == 'check':
		with h5py.File(filename + 'pca.hdf5', 'r') as f:
			ilo = f.attrs['ilo']
			ihi = f.attrs['ihi']
			mean = f['mean'][()]
			V = f['components'][()]
			s = f['values'][()]
			U = f['U'][()]
			lo = f['lo'][()]
			hi = f['hi'][()]
			
			U, s, V, mean = pca_cut(U, s, V, mean, 4)
			z = pca_predict(U, s, V, mean)
			#y = z * (hi - lo).reshape((1,-1)) + lo.reshape((1,-1))
			y = z

			counts = 10**y - 1
			cts = data[data.sum(axis=1)>mincts,ilo:ihi]
			print(counts.shape, cts.shape)
			diff = counts * cts.sum(axis=1).reshape((-1,1)) - cts
			print('largest difference: %.3f' % numpy.abs(diff).max())
			i = numpy.abs(diff).max(axis=1).argmax()
			print(i, 'is the worst case', numpy.abs(diff[i,:]).max())
			iworst = i
			for i in [iworst, 47, 18, 126] + list(range(len(cts))):
				if i >= len(cts): continue
				
				plt.title('ID:%d, %d counts' % (i, cts[i,:].sum()))
				plt.plot(counts[i,:] * cts[i,:].sum(), '-', color='k')
				plt.plot(cts[i,:], '-', color='r', alpha=0.5)
				plt.show()
			
			
if __name__ == '__main__':
	filename = sys.argv[2]
	cmd = sys.argv[1]
	run(cmd=cmd, filename=filename)
